/* @(#)FrIJDEBasicToolBar.java  1.00  2/21/96  Benjamin "Quincy" Cabell V (thethe@wpi.edu), Besiex Software.
 * 
 * Copyright (c) 1996 Besiex Software. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that credit is this copyright notice
 * appears in all copies.  
 *
 * Distributed under the terms of the GNU Library General Public License.
 * 
 * FrIJDE Homepage:
 * http://amber.wpi.edu/~thethe/Document/Besiex/Java/FrIJDE/
 */

package besiex.FrIJDE.AbstractClasses;

import java.applet.*;
import java.awt.*;
import java.awt.image.*;

/**
This is the class that the tool bar must extend.  This class defines the basic requirements of the tool loader
as defined by the FrIJDEMainFrame and the FrIJDETools. <p>


@version 1.00,  21 February 1996

<p>

<b>Changes from last version</b>

<ul>
 	<li> No last version.
</ul>
 
@author <A HREF="http://amber.wpi.edu/~thethe">Benjamin "Quincy" Cabell V</A>
@author <A HREF="http://amber.wpi.edu/~thethe/Documents/Besiex/Java/index.html">Besiex Software</A>
*/

public abstract class FrIJDEBasicToolBar extends FrIJDETool {

	/**
	The actual tool bar window object.  It is to this that tool bar components must be added. <p>
	*/
	
	public FrIJDEBasicToolBarWindow toolBar = null;


	public FrIJDEBasicToolBar () {

	}

	/**
	Sets the FrIJDEMainFrame handle, and creates instance of the tool bar window object. <p>
	
	*@return True if the tool bar was successfully initiated, false otherwise.
	*/

	public boolean Init (FrIJDEMainFrame theFrIJDEFrame) {
		SetFrIJDEMainFrame(theFrIJDEFrame);
		toolBar = new FrIJDEBasicToolBarWindow (theFrIJDEFrame);

		return true;
	}
}



/**
This is the class which FrIJDEBasicToolBar uses to construct the actual tool bar window.  The FrIJDEToolBarWindow
is in fact merely a panel which fits into a Dialog extension (FrIJDEToolBarFrame).  Because the FrIJDEToolBarFrame
has insets, the FrIJDEToolBarWindow must move so as to be placed clear of the insets.  The FrIJDEToolBarFrame
should be of a size just large enough to hold the contents of the FrIJDEToolBarWindow.  The FrIJDEToolBarFrame
must be shown as soon as it is created in order for the insets to be correctly calculated. <p>

<b>Note:</b> This is not an abstract class!  It is a supporting class of an abstract class.<p>


@version 1.00,  21 February 1996

<p>

<b>Changes from last version</b>

<ul>
 	<li> No last version.
</ul>
 
@author <A HREF="http://amber.wpi.edu/~thethe">Benjamin "Quincy" Cabell V</A>
@author <A HREF="http://amber.wpi.edu/~thethe/Documents/Besiex/Java/index.html">Besiex Software</A>
*/


class FrIJDEBasicToolBarWindow extends java.awt.Panel {

	public FrIJDEBasicToolBarFrame panelFrame = null;

	/**
	Constructs the FrIJDEToolBarFrame that will hold this FrIJDEToolBarWindow, show the frame,
	and add this FrIJDEToolBarWindow. <p>
	*/

	public FrIJDEBasicToolBarWindow (FrIJDEMainFrame theFrIJDEMainFrame) {
	
		panelFrame = new FrIJDEBasicToolBarFrame(theFrIJDEMainFrame,"Tool Bar",false);
		show();
		panelFrame.add(this);
	}


	/**
	Show the tool bar. <p>
	*/

	public void show() {
		panelFrame.show();
	}


	/**
	Hide the tool bar. <p>
	*/

	public void hide() {
		panelFrame.hide();
	}

	/**
	Reshapes the FrIJDEToolBarWindow, making sure it is clear of any of its parent insets.  Since the FrIJDEToolBarFrame's
	size is defined as being just large enough to fit the FrIJDEToolBarWindow and its own insets, the FrIJDEToolBarFrame 
	needs to be reshaped as just described. <p>

	*/

	public synchronized void reshape(int x, int y, int width, int height) {
		super.reshape(panelFrame.insets().left,panelFrame.insets().top,width,height);
	
		panelFrame.reshape(panelFrame.location().x,panelFrame.location().y,
				width+panelFrame.insets().left+panelFrame.insets().right,
				height+panelFrame.insets().top+panelFrame.insets().bottom);

	}


	/**
	Returns the preferred size of this FrIJDEToolBarWindow. <p>

	*@return The dimension of a rectangle which contains all the FrIJDEBasicToolBarWindows components,
	 in accordance with the layout manager it uses.
	*/

	public Dimension preferredSize () {
		return (getLayout().preferredLayoutSize(this));
	}

	public Dimension minimumSize () {
		return (preferredSize());
	}

}


/**
This is the class which FrIJDEBasicToolBarWindow uses to construct the actual framed window object.  
It is a dialog, which is essentially a frame which can be made modal or modeless.  In this case, the only behavior this 
tool bar window must handle on its own is the request to destroy itself.  The tool bar window need never actually
destroy itself until FrIJDE is finished, so rather than destroy the tool bar, it is instead simply hidden. <p>

<b>Note:</b> This is not an abstract class!  It is a supporting class of an abstract class.<p>

@version 1.00,  21 February 1996

<p>

<b>Changes from last version</b>

<ul>
 	<li> No last version.
</ul>
 
@author <A HREF="http://amber.wpi.edu/~thethe">Benjamin "Quincy" Cabell V</A>
@author <A HREF="http://amber.wpi.edu/~thethe/Documents/Besiex/Java/index.html">Besiex Software</A>
*/


class FrIJDEBasicToolBarFrame extends java.awt.Dialog {


	public FrIJDEBasicToolBarFrame (FrIJDEMainFrame theFrIJDEMainFrame,String theTitle,boolean theModalBoolean) {
		super(theFrIJDEMainFrame,theTitle,theModalBoolean);
	}

	/**
	Hides the tool bar window when the request is made that it be destroyed.  All other requests are passed
	along to the superclass handler. <p>
	
	*@return True if the event was handled and need not be passed down the handling chain, false otherwise.
	*/

	public boolean handleEvent(Event e) {

		if (e.id == Event.WINDOW_DESTROY) {
			hide();
			return true;
		}

		return super.handleEvent(e);
	}
}
