/* @(#)FrIJDEDocumentManager.java  1.00  2/21/96  Benjamin "Quincy" Cabell V (thethe@wpi.edu), Besiex Software.
 * 
 * Copyright (c) 1996 Besiex Software. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that credit is this copyright notice
 * appears in all copies.  
 *
 * Distributed under the terms of the GNU Library General Public License.
 * 
 * FrIJDE Homepage:
 * http://amber.wpi.edu/~thethe/Document/Besiex/Java/FrIJDE/
 */

package besiex.FrIJDE.AbstractClasses;

import java.awt.*;

/**
Any document manager (multiple document editor) used with FrIJDE must extend this class. 
This class defines all the most basic requirements that other tools would have for 
a multiple document editor.<p>


@version 1.00,  21 February 1996

<p>

<b>Changes from last version</b>

<ul>
 	<li> No last version.
</ul>
 
@author <A HREF="http://amber.wpi.edu/~thethe">Benjamin "Quincy" Cabell V</A>
@author <A HREF="http://amber.wpi.edu/~thethe/Documents/Besiex/Java/index.html">Besiex Software</A>
*/

public abstract class FrIJDEDocumentManager extends FrIJDETool {

		/**
		The document manager creates a new, empty, visible and untitled document.  This document
		appears maximized within the editor-space and is given focus. <p>
		*/

        public abstract boolean NewDocument ();

		/**
		The document manager creates a new, empty, visible document preset with file name and path 
		theFileNameWithPath.  This document appears maximized within the editor-space and is given focus. <p>

		*@param theFileNameWithPath The name and path to give the document created.
		*@return True if successful, false if unsuccessful.
		*/

		public abstract boolean NewDocumentWithFileNameWithPath (String theFileNameWithPath);

		/**
		The document manager opens and makes visible the document with file name and path theFileNameWithPath.  
		This document appears maximized within the editor-space and is given focus. <p>

		*@param theFileNameWithPath The name and path of the file to open.
		*@return True if successful, false if unsuccessful.
		*/

        public abstract boolean OpenDocument (String theFileNameWithPath);


		/**
		The document manager closes (if open) the editor file with file name and path theFileNameWithPath.
		If the document has been modified since last saved (and the document manager is coded with awareness of
		this) the user should be asked if he/she would like to save the document before it is closed).  Focus changes
		only if it is on the document which is to be closed (it should change to the next document in the open document
		list). <p>

		*@param theFileNameWithPath The name and path of the file to close.
		*@return False if unable to close the document (e.g. user cancel from save query), true otherwise.
		*/
		
        public abstract boolean CloseDocument (String theFileNameWithPath);
		

		/**
		The document manager closes all open files (if any).  If any documents have been modified since last saved 
		(and the document manager is coded with awareness of this) the user should be asked if he/she would like to 
		save the documents before they are closed); this should be done via CloseDocument(String).  Focus changes to
		FrIJDEMainFrame. <p>

		*@return False if unable to close any documents (e.g. user cancel from save query), true otherwise.
		*/
		
		public abstract boolean CloseAllDocuments();


		/**
		The document manager opens (if not already opened) the file with the file name and path theFileNameWithPath,
		makes it visible and maximized within the editor-space, gives it focus, jumps to the line number theLineNumber,
		highlights the line, and sets the caret to the beginning of the line. <p>

		*@param theFileNameWithPath The name and path of the file containing the line to go to.
		*@param theLineNumber The number of the line to go to.
		*@return False if unable to open the file or unable to go the line, true otherwise.
		*/

		public abstract boolean GotoLineInDocument (String theFileNameWithPath,int theLineNumber);

		/**
		The document manager opens (if not already opened) the file with the file name and path theFileNameWithPath,
		makes it visible and maximized within the editor-space, gives it focus, jumps to the line number theLineNumber,
		highlights the line, and sets the caret at character position theCharacterPosition. <p>

		*@param theFileNameWithPath The name and path of the file containing the line to go to.
		*@param theLineNumber The number of the line to go to.
		*@param theCharacterPosition The character position (number of characters from the beginning of the line)
				at which to place the caret.
		*@return False if unable to open the file or unable to go the line, true otherwise.
		*/

		public abstract boolean GotoLineInDocument (String theFileNameWithPath,int theLineNumber,
				int theCharacterPosition);

}
