/* @(#)FrIJDEProcessRunnerForTool.java  1.00  2/24/96  Benjamin "Quincy" Cabell V (thethe@wpi.edu), Besiex Software.
 * 
 * Copyright (c) 1996 Besiex Software. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that credit is this copyright notice
 * appears in all copies.  
 *
 * Distributed under the terms of the GNU Library General Public License.
 * 
 * FrIJDE Homepage:
 * http://amber.wpi.edu/~thethe/Document/Besiex/Java/FrIJDE/
 */

package besiex.FrIJDE.AbstractClasses;

import awtExt.*;
import sysExt.*;
import java.io.*;

/**
Some FrIJDETools need to run executable code and capture their output; these FrIJDETools can act as interfaces for
previously developed code such as Sun's Javac, Java, etc.  This class makes it easier for these kinds
of FrIJDETools to be developed. <p>


@version 1.00,  24 February 1996

<p>

<b>Changes from last version</b>

<ul>
 	<li> No last version.
</ul>
 
@author <A HREF="http://amber.wpi.edu/~thethe">Benjamin "Quincy" Cabell V</A>
@author <A HREF="http://amber.wpi.edu/~thethe/Documents/Besiex/Java/index.html">Besiex Software</A>
*/


public abstract class FrIJDEProcessRunnerForTool extends Thread {

	/**
	Parent FrIJDETool.
	*/

	public FrIJDETool parent;

	/**
	The command line to use.
	*/
	String commandLine;

	/**
	The handle of the process.
	*/
	public Process runningProcess;

	DataInputStream processOutput;
	DataOutputStream processInput;
	
	/**
	The constructor used to create the process to run. <p>
	
	*@param theParent The FrIJDETool which instantiates this class.
	*@param theCommandLine The String to use to execute the process.
	*/

	public FrIJDEProcessRunnerForTool(FrIJDETool theParent, String theCommandLine) {
		parent = theParent;
		commandLine = theCommandLine;
		run();
	}


	/**
	This method is called every time the process this class runs returns a line of output.  The response to
	that output is returned by this method and passed into the process as input. <p>
	
	*@return The String to pass as input to the process.
	*/

	public abstract String ProcessLineOutputHandler (String theProcessOutputAsString);


	/**
	When the process is started, this method is called.
	*/
	
	public abstract void ProcessStarted ();


	/**
	When the process is ended, this method is called.
	*/

	public abstract void ProcessEnded ();


	/**
	This method is executed by the constructor.
	*/

	public void run() {
		try {
			runningProcess = Runtime.getRuntime().exec(commandLine);
			processOutput = new DataInputStream(new BufferedInputStream(runningProcess.getInputStream()));
	  		processInput = new DataOutputStream(new BufferedOutputStream(runningProcess.getOutputStream()));
	  
			ProcessStarted();
			String outputOfProcessAsString = null;
			String inputOfProcessAsString = null;
			
			while ((outputOfProcessAsString = processOutput.readLine()) != null) {
				inputOfProcessAsString = ProcessLineOutputHandler(outputOfProcessAsString);
				if (inputOfProcessAsString != null)
					processInput.writeChars(inputOfProcessAsString);
			}
		} catch (Exception e) {
			 System.out.println("FrIJDEProcessRunnerForToolRead::run() -- Exception occurred " + e);
		}
		ProcessEnded();
	}
}

