/* @(#)MightyButtonWithCallback.java  1.80  2/21/96  Benjamin "Quincy" Cabell V (thethe@wpi.edu), Besiex Software.
 * 
 * Copyright (c) 1996 Besiex Software. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that credit is this copyright notice
 * appears in all copies.  Blah, blah, blah. Blah, blah.
 */
 
import java.awt.*;
import java.applet.*;
import java.util.*;


import awtExt.*;
import sysExt.*;

/**    
 The MightyButtonWithCallback class is used to create various types of buttons, 
 MOMENTARY, TOGGLE, PERMANENT, or RADIO.  The appearance of the MightyButtonWithCallback
 in its states (PRESSED, NOT_PRESSED, and DISABLED) is determined by the
 <code>buttonName</code> passed to the constructor (this name is used to
 load in the 3 GIFs).  See constructor for details. <p>
 
 Using the MightyRadioButtonGroup class supplied, any MightyButtonWithCallback can be a 
 radio button. <p>
 
 Button Types:<br>
 <UL>
 	<LI>A MOMENTARY button will stay PRESSED only as long as the mouse is down over it.
 	<LI>A TOGGLE button will become PRESSED if NOT_PRESSED, and NOT_PRESSED if PRESSED when the mouse
 	button is pressed over the MightyButtonWithCallback.
 	<LI>A PERMANENT button will stay PRESSED forever, once PRESSED.
 	<LI>A RADIO button is used with the MightyRadioButtonGroup for radio button functionality.
 </UL><P>
 
 Behavior is added to MightyButtonWithCallback by using <b>Sal Cataudella's (rangerx@panix.com)</b> 
 <a href="http://www.panix.com/~rangerx/packages.html">awtExt and sysExt</a>.
 
@version 1.80,  21 February 1996
 
<p>
<b>Changes from last version</b>

<ul>
 	<li> Automatically detects whether the MightyButton is in a stand-alone application or an applet and accesses the images
		accordingly.
</ul>
 
@author <A HREF="http://amber.wpi.edu/~thethe">Benjamin "Quincy" Cabell V</A>
@author <A HREF="http://amber.wpi.edu/~thethe/Documents/Besiex/Java/index.html">Besiex Software</A>
@author built from ImageButton by <A HREF="http://www.cs.brown.edu/people/amd/">Adam Doppelt</A>

 */

public class MightyButtonWithCallback extends MightyButton {

	//This button's boolean states 
	boolean actionCanBeCalled=true;
	boolean hasBeenPressed=false;
	boolean buttonPressedOnLastMouseDown=false;
	
	methodRef actionMethodRef = null ;



/**
 * Construct an image button. An image button is composed of three
 * images, one for each possible state of the button. If null is
 * passed for pressed and/or disabled, those states will draw using the
 * normal image.  Many thanks to Bobby Arnold (rharnold@csupomona.edu) for
 * his recommendation on getting the Applet's handle from within a component. <p>
 *
 * @param theButtonName The file name of the button, without the _p.gif,_n.gif,_d.gif.  
 * Here's how it works:  I tack on "_p.gif" for the image of a pressed button,
 * "_n.gif" for the image of a not-pressed button, and "_d.gif" for the image of a 
 * disabled button.  So, if you pass in "iAmAButton" as theButtonName, this
 * code will load in iAmAButton_p.gif, iAmAButton_n.gif, and iAmAButton_d.gif.
 * Get it?  It was meant to make things easy.  You also have to specify the location of the
 * button image file here.  In the case of a stand-alone app this must be the absolute path
 * to the file, in the case of an applet, this is relative to the codebase.<p>
 *
 * @param theButtonText When the button is displayed, you can have it displayed with
 * text on top of it (so people know what the button does).  Text color, font, and
 * size can all be set (default color is black, default font is Helvetica Bold 12). <p>
 *
 * @param theButtonType What kind of button is this?  MOMENTARY, TOGGLE, PERMANENT, or
 * RADIO.
 * */

	public MightyButtonWithCallback (String theButtonName,String theButtonText,
			int theButtonType) {
		this(null,theButtonName,theButtonText,theButtonType,null);
	}


/**
 * Constructor for a MOMENTARY MightyButtonWithCallback with no text put onto the button. <P>
 *
 * @param theApplet A handle to the Applet; This is required for using 
 * getDocumentBase(), getCodeBase(), and getImage(URL, String).  There should be,
 * nay, must be a way to get the Applet handle without having to pass it in,
 * but I couldn't figure one out.  Anyone know if it can be done (and how)?<P>
 */

	public MightyButtonWithCallback (Applet theApplet,String theButtonName) {
		this(theApplet,theButtonName,null,MOMENTARY,null);
	}


/**
*
*Constructor for giving a target and method name for defining callback behavior.  See awtExt and sysExt for details. <p>
*
* @param theTarget The object (instance of the class) in which the callback method will be found.  (See Sal Cataudella's (rangerx@panix.com) 
* 	<a href="http://www.panix.com/~rangerx/packages.html">awtExt and sysExt</a> for details).
* @param theMethodName The name of the method that should be called back.  (See Sal Cataudella's (rangerx@panix.com) 
* 	<a href="http://www.panix.com/~rangerx/packages.html">awtExt and sysExt</a> for details).
* 
*/

	public MightyButtonWithCallback (String theButtonName,String theButtonText, int theButtonType,
			Object theTarget, String theMethodName) throws InvalidMethodRefException {
		this (null,theButtonName,theButtonText,theButtonType,Callback.newRef(theTarget,theMethodName));
	}


/**
*
*Constructor for use with methodRef. See awtExt and sysExt for details. <p>
*
* @param theMethodRef The methodRef to callback for this button.  (See Sal Cataudella's (rangerx@panix.com) 
* 	<a href="http://www.panix.com/~rangerx/packages.html">awtExt and sysExt</a> for details).
* 
*/

	public MightyButtonWithCallback (String theButtonName,String theButtonText, int theButtonType, methodRef theMethodRef) {
		this (null,theButtonName,theButtonText,theButtonType,theMethodRef);
	}

	
	public MightyButtonWithCallback (Applet theApplet,String theButtonName,String theButtonText, int theButtonType,
			methodRef theMethodRef) {
		super (theApplet,theButtonName,theButtonText,theButtonType);

		actionMethodRef = theMethodRef;
	}


	public MightyButtonWithCallback (Applet theApplet,String theButtonName,String theButtonText, int theButtonType,
			Object theTarget, String theMethodName) throws InvalidMethodRefException {
		this (theApplet,theButtonName,theButtonText,theButtonType,Callback.newRef(theTarget,theMethodName));
	}




/**
 * INTERNAL USE ONLY.<BR>
 * This method implements the button's type.  When the button is released, this
 * method gets called for changing the button's state (if appropriate to the 
 * button's type).
 */

	void buttonReact() {

		switch (buttonType) {

			case (MOMENTARY)	:		SetState(NOT_PRESSED);
										break;
		
			case (RADIO) 		:		MightyButtonWithCallback c;
									
											for (int i=0;i<buttonsForRadioPress.size();i++) {
												c = (MightyButtonWithCallback) buttonsForRadioPress.elementAt(i);
												c.SetState(NOT_PRESSED);
											}

											SetState(PRESSED);
											break;
											
			case (PERMANENT) 	: 		break;
	
			case (TOGGLE)		:		if (!buttonPressedOnLastMouseDown) {
												SetState(NOT_PRESSED);
											}
											break;
		}
	}



/**
* Adds user defined methodRef. This is invoked when a
* ACTION_EVENT occurs. The methodRef must be
* defined using the utility class awtExt.Callback
* @param aMethodRef ACTION_EVENT event handling method.
* must have the target object set.
* @see sysExt.methodRef#setTarget
* @see awtExt.Callback
* @see sysExt.methodRef
* 
**/
	public void	addActionCallback( methodRef aMethodRef ) {
		actionMethodRef = aMethodRef ;
	}



/**
* Called when an action occurs. This in turn calls back
* a user defined method on a target object. The called back
* method *AND* the target object must have a specific signature.
* Since the action event for this button is *always* handled 
* here, this action() method returns false to stop the event
* propagation.
* @param evt The action event that just occurred.
* @param obj The label of this object.
* @see 	awtExt.Callback
* @see	awtExt.CallbackInfo
**/

	public boolean action(Event evt,Object obj ) {
		buttonReact();
				
		if (actionCanBeCalled) {
			if (buttonType == PERMANENT)
				actionCanBeCalled = false;

			if (action())
				return true;
			else {
				// This event was NOT handled by action(), so allow it to be handled here.
				 CallbackInfo cbInfo = new CallbackInfo ( evt, buttonText, this ) ;

				if(actionMethodRef != null ) {
					actionMethodRef.doAction( cbInfo );

					if (cbInfo.processed)
						return true;
					else
						return false;
				}
			} 
				
		} else
			return true;

		return false;
	}



/**
 * INTERNAL USE ONLY.<BR>
 * Mouse down causes the button to be pressed if the button is enabled. <p>
 * @param event The Event that has occured. <p>
 * @param x The x coordinate of the event. <p>
 * @param y The y coordinate of the event.
 */

	public boolean mouseDown(Event event, int x, int y) {
		if (((buttonType == PERMANENT) && (!hasBeenPressed)) 
				|| ((buttonType != PERMANENT) && (buttonState != PRESSED) 
				&& (buttonState != DISABLED))) {

			SetState(PRESSED);
		
			buttonPressedOnLastMouseDown=true;
			hasBeenPressed = true;

		} else
			buttonPressedOnLastMouseDown = false;
		
		return true;
	}


/**
 * INTERNAL USE ONLY.<BR>
 * When the mouse button goes up, the button reacts, and the action is called. <p>
 * @param event The Event that has occured. <p>
 * @param x The x coordinate of the event. <p>
 * @param y The y coordinate of the event.
 */
	public boolean mouseUp(Event theEvent, int x, int y) {

		Event event = theEvent;
		event.id = Event.ACTION_EVENT;
		return (postEvent(event));
	}

}


