/* @(#)MightyRelativeLayout.java  2.20  2/21/96  Benjamin "Quincy" Cabell V (thethe@wpi.edu), Besiex Software.
 * 
 * Copyright (c) 1996 Besiex Software. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that credit is this copyright notice
 * appears in all copies.  Blah, blah, blah. Blah, blah.
 */
 
import java.awt.* ;
import java.util.* ;
import java.lang.* ;

/**
The MightyRelativeLayout class is used to layout java.awt.Components relative to
each other.  Requires <a href="MightyRelativeConstraint.html">MightyRelativeConstraint</a>.
<p>

<a name="usage"></a>

<dl>
	<dt><i>Usage:</i></dt>
	<dd>add("name of component;relative to=none;placement=base;dx=0;dy=0;",handleOfComponent);
	<dd>add("name of component;relative to=none;placement=base;",handleOfComponent);
	<dd>add("name of component;relative to=none;placement=base;anchor=bottom", handleOfComponent);
	<dd>add("name of component;relative to=previously added component's name;placement=upper right corner is middle", handleOfAnotherComponent);
</dl>

<b>Parameter (<i>name of component</i>)</b><br>
<ul>
	<li> a uniquely identifying component name that will be used if other components
		  are to be placed relative to this one (the name can be anything)
</ul>

<b>Parameter <i>relative to=</i></b><br>
<ul>
	<li>none
	<li><b>the uniquely identifying component name of a previously added component</b>
</ul>

<b>NOTE:</b> A component should *never* be added to the MightyRelativeLayout before the component
to which it is placed relative.<p>

<b>Parameter <i>placement=</i></b><br>
<ul>
	<li><i>base</i> - the first component added to the MightyRelativeLayout is the base
	<li><i>above</i> - centered above the reference component
	<li><i>under</i> - centered under the reference component 
	<li><i>below</i> - centered below the reference component (exactly the same as under)
	<li><i>left</i> - centered above the reference component 
	<li><i>right</i> - centered above the reference component 

	<li><i>left anchored</i> - placed along the horizontal anchor, to the left of the reference component
	<li><i>right anchored</i> - placed along the horizontal anchor, to the right of the reference component
	<li><i>top anchored</i> - placed along the vertical anchor, above the reference component
	<li><i>bottom anchored</i> - placed along the vertical anchor, below the reference component

	<li><i>upper right corner is middle</i> - upper right corner of reference component is the middle left
	<li><i>lower right corner is middle</i> - lower right corner of reference component is the middle left

	<li><i>above right</i> - right-aligned above the reference component 
	<li><i>under right</i> - righ-aligned under the reference component 
	<li><i>below right</i> - righ-aligned below the reference component (same as under right)
	<li><i>left bottom</i> - centered above the reference component 
	<li><i>right bottom</i> - centered above the reference component 
	<li><i>upper left corner</i> - upper left at upper left of reference component
	<li><i>upper right corner</i> - upper right at upper right of reference component
	<li><i>lower right corner</i> - lower right at lower right of reference component
	<li><i>lower left corner</i> - lower left at lower left of reference component
	<li><i>inside upper left</i> - upper left at the upper left of reference component
	<li><i>inside upper right</i> - upper right at the upper right of reference component
	<li><i>inside lower right</i> - lower right at the lower right of reference component
	<li><i>inside lower left</i> - lower left at the lower left of reference component
	<li><i>above left</i> - left-aligned above the reference component 
	<li><i>under left</i> - left-aligned under the reference component 
	<li><i>below left</i> - left-aligned below the reference component 
	<li><i>left top</i> - top-aligned to the left of the reference component 
	<li><i>right top</i> - top-aligned to the right of the reference component 
	<li><i>inside top</i> - centered with the top at the top of the reference component
	<li><i>inside bottom</i> - centered with the bottom at the bottom of the reference component
	<li><i>inside right</i> - centered with the right at the right of the reference component
	<li><i>inside left</i> - centered with the left at the left of the reference component
</ul>

<b>Parameter <i>anchor=</i></b><br>
<ul>
	<li>center - specifies both vertical and horizontal anchors at center of component
	<li>top - specifies vertical anchor at top of component
	<li>bottom - specifies vertical anchor at bottom of component
	<li>left - specifies horizontal anchor at left of component
	<li>right - specifies horizontal anchor at right of component
	<li>adopt - forces this component (which must be a container) to adopt the horizontal and vertical anchors 
		of a child component (the search for the child component starts at the base and ends when an anchor is
		specified)
</ul>

 
@version 2.20,  21 February 1996
<p>

<b>Changes from last version</b>

<ul>
 	<li> Now allows horizontal and vertical anchors which can be contained within containers that are to be laid out.
		The most obvious use of this would be for aligning items which are subscripted and superscripted in the normal
		manner (which means aligning all the items along the baseline of the non-sub/superscripted items.  I can't
		think of any better way to describe this feature.
</ul>
 
@author <A HREF="http://amber.wpi.edu/~thethe">Benjamin "Quincy" Cabell V</A>
@author <A HREF="http://amber.wpi.edu/~thethe/Documents/Besiex/Java/index.html">Besiex Software</A>
@author built from Jean-Claude Dufourd's <a href="http://www-elec.enst.fr/java/RelativeLayout.java">RelativeLayout
			Manager</a>

*/

public class MightyRelativeLayout implements LayoutManager {

	// Tags for Anchor
	final static String tagAdopt = "adopt";
	final static String tagBottom = "bottom";
	final static String tagTop = "top";
	final static String tagCenter = "center";

	// Tags for anchored placement

	final static String tagRightAnchored = "right anchored";
	final static String tagLeftAnchored = "left anchored";
	final static String tagTopAnchored = "top anchored";
	final static String tagBottomAnchored = "bottom anchored";


	// Tags for normal (non-anchored) placement

	final static String tagBase = "base";
	final static String tagAbove = "above";
	final static String tagUnder = "under";
	final static String tagBelow = "below";
	final static String tagLeft = "left";
	final static String tagRight = "right";
	final static String tagAboveRight = "above right";
	final static String tagUnderRight = "under right";
	final static String tagBelowRight = "below right";
	final static String tagLeftBottom = "left bottom";
	final static String tagRightBottom = "right bottom";
	final static String tagUpperLeftCorner = "upper left corner";
	final static String tagUpperRightCorner = "upper right corner";
	final static String tagLowerRightCorner = "lower right corner";
	
	final static String tagUpperRightCornerIsMiddle = "upper right corner is middle";
	final static String tagLowerRightCornerIsMiddle = "lower right corner is middle";

	final static String tagLowerRightCornerIsQuarterBelowMiddle = "lower right corner is quarter below middle";
	final static String tagLowerLeftCornerIsSixthBelowMiddle = "lower right corner is sixth below middle";
	final static String tagLowerRightCornerIsSixthBelowMiddle = "lower right corner is sixth below middle";

	final static String tagLowerLeftCorner = "lower left corner";
	final static String tagInsideUpperLeft = "inside upper left";
	final static String tagInsideUpperRight = "inside upper right";
	final static String tagInsideLowerRight = "inside lower right";
	final static String tagInsideLowerLeft = "inside lower left";
	final static String tagAboveLeft = "above left";
	final static String tagUnderLeft = "under left";
	final static String tagBelowLeft = "below left";
	final static String tagLeftTop = "left top";
	final static String tagRightTop = "right top";
	final static String tagInsideTop = "inside top";
	final static String tagInsideBottom = "inside bottom";
	final static String tagInsideRight = "inside right";
	final static String tagInsideLeft = "inside left";

	final static String tagNone = "none";

	final static String tagRelativeTo = "relative to";
	final static String tagPlacement	= "placement";
	final static String tagAnchor = "anchor";
	final static String tagDx = "dx";
	final static String tagDy = "dy";

	final static String CHAR_SEMI = ";";
	final static String CHAR_EQUAL = "=";
	final static String CHAR_ALL = CHAR_SEMI + CHAR_EQUAL;

	int width=0;
	int height=0;
	
	int windows95FixRepaint = 2;

	Dimension layoutDimensions = new Dimension(0,0);
	Vector components = new Vector(10,10);
	Vector constraints = new Vector(10,10);

	int ignoreCount=0;


	public int hAnchorOffset=0;
	public int vAnchorOffset=0;

	public MightyRelativeLayout() {

	}


/**
AUTOMATIC USE ONLY.<P>

Add theComponent to the MightyRelativeLayout.  The standard method promoted through
the design of the java.awt.Container.add(String,Component) is used.  The String,
theParameters, contains the information as outlined <a href="#usage">above</a>.<p>

Required by <b>java.awt.LayoutManager</b>.<P>

@param theParameters The parameters used to add theComponent.  <a href="#usage">
		 					Details</a>.
@param theComponent The java.awt.Component to add to the MightyRelativeLayout 
						  manager.

*/

	public void addLayoutComponent(String theParameters, Component theComponent) {

		String name = null;
		String reference = null;
		String placement = null;
		String tempDx="0"; 
		String tempDy="0";	
		String tok, val;
		String tempAnchor=null;
		int dx=0;
		int dy=0;

		try {
			if (theComponent == null) {
				System.out.println("MightyRelativeLayout::addLayoutComponent -- Error!  Null Component given to lay out!  Parameters=\""+theParameters+"\"");
				System.exit(1);
			}
			else {
				StringTokenizer st = new StringTokenizer(theParameters, CHAR_ALL, true);

				name = new String(st.nextToken());

				while(st.hasMoreTokens()) {
					if (st.nextToken().equals(CHAR_SEMI)) {

					   if (st.hasMoreTokens()) {
							tok = st.nextToken();
					
							if (!st.nextToken().equals(CHAR_EQUAL))
								throw new NoSuchElementException();

							if (tok.equals(tagRelativeTo)) {
								reference = st.nextToken();
							} 
							else 
								if (tok.equals(tagPlacement)) {
									placement = st.nextToken();
								}
								else 
									if (tok.equals(tagDx)) {
										tempDx = st.nextToken();
									}
									else 
										if (tok.equals(tagDy)) {
											tempDy = st.nextToken();
										}
										else 
											if (tok.equals(tagAnchor)) {
												tempAnchor = st.nextToken();
											}

						}
					}
				}
			}

		} catch (Exception e) {

			System.out.println("MightyRelativeLayout::addLayoutComponent -- Error!  Syntax error in parameters!  Parameters=\""+theParameters+"\"");
			System.exit(1);
    	}

		Component referenceComponent = GetComponentFromName(reference);
		
		if (referenceComponent == null) {
			if (reference != null) {
				if (!reference.equals ("none")) {
					System.out.println ("MightyRelativeLayout::addLayoutComponent -- Error!  Cannot find the relative component \""+reference+"\" that is referred to by \""+name+"\"");
					System.exit(1);
				}
			}
			else {
				System.out.println ("MightyRelativeLayout::addLayoutComponent -- Error!  Cannot find the relative to specifier for \""+name+"\"");

				System.out.println("Name: "+name+" Ref: "+reference+" placement: "+placement+" dx:"+tempDx+" dy:"+tempDy);
				System.exit(1);
			}

		}

		dx = Integer.valueOf(tempDx).intValue();
		dy = Integer.valueOf(tempDy).intValue();

		MightyRelativeConstraint c = new MightyRelativeConstraint (name, referenceComponent, 
				placement, dx, dy, tempAnchor);

		components.addElement( theComponent);
		constraints.addElement( c );
	}


/**
INTERNAL USE ONLY.<P>

Returns the java.awt.Component previously added to the MightyRelativeLayout with 
theName; null is returned if no component has theName.

@param theName The name of the java.awt.Component to get.
@return The java.awt.Component previously added to the MightyRelativeLayout with 
			theName; null is returned if no component has theName.
*/
	Component GetComponentFromName(String theName) {

		Component componentInLayout;
		MightyRelativeConstraint componentLayoutConstraint;

		for (int i=0; i<components.size() ;i++) { 
			componentInLayout = (Component) components.elementAt(i);
			componentLayoutConstraint = (MightyRelativeConstraint) constraints.elementAt(i);
			
			if (componentLayoutConstraint.name.equals(theName)) {
				return componentInLayout;
			}
		}
		return null;
	}


/**
AUTOMATIC USE ONLY<P>

Remove theComponent from the MightyRelativeLayout.<p>

Required by <b>java.awt.LayoutManager</b>.<P>

@param theComponent The java.awt.Component to remove from the MightyRelativeLayout 
						  manager.
*/

	public void removeLayoutComponent(Component theComponent) {
		int i = components.indexOf(theComponent);
		if (i != -1) {
			components.removeElementAt( i );
			constraints.removeElementAt( i );
		}
	}


/**
AUTOMATIC USE ONLY.<P>

@param theContainer The container which holds the components.<p>

Required by <b>java.awt.LayoutManager</b>.<P>

@return The minimum Dimension required to display all the components in the
			MightyRelativeLayout manager.
*/

	public Dimension minimumLayoutSize(Container theContainer) {
		layoutContainer(theContainer);
		return new Dimension (width,height);
	}


/**
AUTOMATIC USE ONLY.<P>

In the case of MightyRelativeLayout, the preferred layout size returned is in 
fact the minimum layout size.<p>

Required by <b>java.awt.LayoutManager</b>.<P>

@param theContainer The container which holds the components.
@return The preferred Dimension to use in displaying all the components in the
			MightyRelativeLayout manager.  
*/
    
	public Dimension preferredLayoutSize(Container theContainer) {
		return minimumLayoutSize(theContainer);
	}


/**
INTERNAL USE ONLY.<P>

The minimum Dimension which includes all components in the layout manager requires
keeping track of how the layout size must change to accomodate new components.  This
method determines whether or not the added component (which has the bounds passed in)
breaks the current bounds of the layout.  If it does, the width and height are
readjusted.
*/
	
	void CountForLayoutBoundary(int theComponentIndex) {

		Component currentComponent = (Component) components.elementAt(theComponentIndex);
	
		int currentX = currentComponent.bounds().x;		
		int currentY = currentComponent.bounds().y;
		int currentWidth = currentComponent.bounds().width;		
		int currentHeight = currentComponent.bounds().height;
	
		int incomingMaxX = currentX + currentWidth;
		int incomingMaxY = currentY + currentHeight;
		
		int incomingMinX = currentX;
		int incomingMinY = currentY;
		
		if (incomingMaxX > width)
			width = incomingMaxX;
			
		if (incomingMinX < 0)
			width += -incomingMinX;
		
		if (incomingMaxY > height)
			height = incomingMaxY;

		if (incomingMinY < 0)
			height += -incomingMinY;
	}

	
/**
INTERNAL USE ONLY.<P>

Readjust all components that have already been positioned to account for the change
in layout size (which shifts all already placed components). <p>

<b>NOTE:</b>	There is a bit of a hack here.  When each already laid out component is
					moved, via java.awt.Component.move(int,int) (the only way I can figure out
					to move a component), layoutContainer gets called.  This creates an infinite
					loop.  So, when a component is moved, the layoutContainer (Container) function
					will do nothing on its next two calls (thus avoiding the infinite loop).<p>
*/

	void AdjustContainerUpTo(int theHighInclusiveComponentNumber,int theXOffset,int theYOffset) {

		Component componentInLayout;

		for (int i=0; i<= theHighInclusiveComponentNumber ;i++) {  
			componentInLayout = (Component) components.elementAt(i);
			
			ignoreCount = 2;
			componentInLayout.move(componentInLayout.bounds().x+theXOffset,
					componentInLayout.bounds().y+theYOffset);
			componentInLayout.invalidate();
		}
	}

/**
INTERNAL USE ONLY.<P>

If the component under analysis, the one with the index theComponentIndex, is placed 
above or to the left of the current bounds of the layout, then the layout must be 
resized and all previously laid out components shifted so that they retain their 
relative positions.  This method checks to see if the component (with theComponentIndex)
forces the layout to be resized and if so, it calls for AdjustContainerUpTo with the
index of the highest laid out component (which is this one).

@param theComponentIndex The index of the component to adjust for, if necessary.
*/

	void AdjustIfNecessary (int theComponentIndex) {
		
		Component currentComponent = (Component) components.elementAt(theComponentIndex);

		int currentX = currentComponent.bounds().x;		
		int currentY = currentComponent.bounds().y;

		int xOffset = 0;
		int yOffset = 0;
		
		if (currentX < 0)
			xOffset = - currentX;
			
		if (currentY < 0)
			yOffset = - currentY;

	
		if ((xOffset != 0) || (yOffset != 0)) {

			if (hAnchorOffset != -1)
				hAnchorOffset += xOffset;

			if (vAnchorOffset != -1)
				vAnchorOffset += yOffset;

			AdjustContainerUpTo(theComponentIndex,xOffset,yOffset);
		}
	


	}


	void AnchorProcess (int theComponentIndex) {
		Component currentComponent = (Component) components.elementAt(theComponentIndex);
		MightyRelativeConstraint currentConstraint = (MightyRelativeConstraint) constraints.elementAt(theComponentIndex);

		if (currentConstraint.anchor != null) {
			if (currentConstraint.anchor.equals(tagAdopt)) {
				// Set the *this* layout manager's anchor to the anchor of this component.

				MightyRelativeLayout currentMRL = GetMRLOfComponent (currentComponent);

				if (currentMRL != null) {

					hAnchorOffset = currentComponent.bounds().x+currentMRL.hAnchorOffset;
					vAnchorOffset = currentComponent.bounds().y+currentMRL.vAnchorOffset;

				} else {
					System.out.println("MightyRelativeLayout::AnchorProcess -- Major problem, tring to adopt from a component that does not use MightyRelativeLayout");
					System.exit(1);
				}
		
			} 
			else
			if (currentConstraint.anchor.equals(tagBottom)) {
				hAnchorOffset = -1; 
				vAnchorOffset = currentComponent.bounds().y+currentComponent.bounds().height;
			}
			else
			if (currentConstraint.anchor.equals(tagCenter)) {
				hAnchorOffset = currentComponent.bounds().x+(int) (currentComponent.bounds().width/2); 
				vAnchorOffset = currentComponent.bounds().y+(int) (currentComponent.bounds().height/2);
			}
			else {
				System.out.println("MightyRelativeLayout::AnchorProcess -- Not supporting that anchoring yet!");
				System.exit(1);
			}
		}
	}

		
	
/**
AUTOMATIC USE ONLY.<P>
	
The java.awt.LayoutManager requires this.  This method is automatically called by
the Container which has been set with this instance of the MightyRelativeLayout.
This method is also called when minimumLayoutSize or preferredLayoutSize is called;
this is necessary to calculate the size of the layout.<p>

Required by <b>java.awt.LayoutManager</b>.<P>

@param theContainer The container which holds the components.
*/
	public synchronized void layoutContainer(Container theContainer) {

		if (ignoreCount > 0) {
			ignoreCount -= 1;		
//			System.out.println("Withholding layout");
		} else {

			width=0;
			height=0;
			
			Rectangle componentBounds, referenceComponentBounds;
			MightyRelativeConstraint componentLayoutConstraint;
			Dimension componentPreferredSize;
			Component componentInLayout;
			
			for (int i=0;i < components.size();i++) {

				componentLayoutConstraint = (MightyRelativeConstraint) constraints.elementAt(i);
				componentInLayout = (Component) components.elementAt(i);
	
				componentPreferredSize = componentInLayout.preferredSize();
				componentBounds = componentInLayout.bounds();
			
				if (componentLayoutConstraint.reference == null)
					referenceComponentBounds = theContainer.bounds();
				else 
					referenceComponentBounds = componentLayoutConstraint.reference.bounds();
	
				int layoutToX=0;
				int layoutToY=0;
				int layoutToWidth=0;
				int layoutToHeight=0;

				if (componentLayoutConstraint.placement.equals(tagAbove)) {
					layoutToX = referenceComponentBounds.x + referenceComponentBounds.width/2 - componentBounds.width/2 + componentLayoutConstraint.dx;
					layoutToY = referenceComponentBounds.y - componentLayoutConstraint.dy - componentBounds.height;
				}
				else
				if (componentLayoutConstraint.placement.equals(tagUnder) ||
						componentLayoutConstraint.placement.equals(tagBelow)) {

					layoutToX =  referenceComponentBounds.x + referenceComponentBounds.width/2 - componentBounds.width/2 + componentLayoutConstraint.dx;
					layoutToY = referenceComponentBounds.y + componentLayoutConstraint.dy + referenceComponentBounds.height;
				}
				else
				if (componentLayoutConstraint.placement.equals(tagLeft)) {
					layoutToX = referenceComponentBounds.x - componentLayoutConstraint.dx - componentBounds.width;
					layoutToY = referenceComponentBounds.y + referenceComponentBounds.height/2 - componentBounds.height/2 + componentLayoutConstraint.dy;
				}
				else
				if (componentLayoutConstraint.placement.equals(tagRight)) {
					layoutToX = referenceComponentBounds.x + componentLayoutConstraint.dx + referenceComponentBounds.width;
					layoutToY = referenceComponentBounds.y + referenceComponentBounds.height/2 - componentBounds.height/2 + componentLayoutConstraint.dy;
				}
				else		
				if (componentLayoutConstraint.placement.equals(tagAboveRight)) {
					layoutToX = referenceComponentBounds.x + referenceComponentBounds.width - componentBounds.width - componentLayoutConstraint.dx;
					layoutToY = referenceComponentBounds.y - componentLayoutConstraint.dy - componentBounds.height;
				}	
				else
				if (componentLayoutConstraint.placement.equals(tagUnderRight) ||
						componentLayoutConstraint.placement.equals(tagBelowRight)) {
					layoutToX = referenceComponentBounds.x + referenceComponentBounds.width - componentBounds.width - componentLayoutConstraint.dx;
					layoutToY = referenceComponentBounds.y + componentLayoutConstraint.dy + referenceComponentBounds.height;
				}
				else
				if (componentLayoutConstraint.placement.equals(tagLeftBottom)) {
					layoutToX = referenceComponentBounds.x - componentLayoutConstraint.dx - componentBounds.width;
					layoutToY = referenceComponentBounds.y + referenceComponentBounds.height - componentBounds.height - componentLayoutConstraint.dy;
				}
				else
				if (componentLayoutConstraint.placement.equals(tagRightBottom)) {
					layoutToX = referenceComponentBounds.x + componentLayoutConstraint.dx + referenceComponentBounds.width;
					layoutToY = referenceComponentBounds.y + referenceComponentBounds.height - componentBounds.height - componentLayoutConstraint.dy;
				}
				else
				if (componentLayoutConstraint.placement.equals(tagUpperLeftCorner)) {
					layoutToX = referenceComponentBounds.x - componentLayoutConstraint.dx - componentBounds.width;
					layoutToY = referenceComponentBounds.y - componentLayoutConstraint.dy - componentBounds.height;
				}
				else
				if (componentLayoutConstraint.placement.equals(tagUpperRightCorner)) {
					layoutToX = referenceComponentBounds.x + referenceComponentBounds.width + componentLayoutConstraint.dx;
					layoutToY = referenceComponentBounds.y - componentLayoutConstraint.dy - componentBounds.height;
				}
				else
				if (componentLayoutConstraint.placement.equals(tagLowerRightCorner)) {
					layoutToX = referenceComponentBounds.x + referenceComponentBounds.width + componentLayoutConstraint.dx;
					layoutToY = referenceComponentBounds.y + componentLayoutConstraint.dy + referenceComponentBounds.height;
				}
				else
				if (componentLayoutConstraint.placement.equals(tagLowerRightCornerIsMiddle)) {
					layoutToX = referenceComponentBounds.x + referenceComponentBounds.width + componentLayoutConstraint.dx;
					layoutToY = referenceComponentBounds.y + componentLayoutConstraint.dy 
							+ referenceComponentBounds.height - (int) (componentBounds.height/2);
				}
				else
				if (componentLayoutConstraint.placement.equals(tagLowerRightCornerIsQuarterBelowMiddle)) {
					layoutToX = referenceComponentBounds.x + referenceComponentBounds.width + componentLayoutConstraint.dx;
					layoutToY = referenceComponentBounds.y + componentLayoutConstraint.dy
							+ referenceComponentBounds.height - (int) ((3/4)*componentBounds.height);
				}
				else
				if (componentLayoutConstraint.placement.equals(tagLowerRightCornerIsSixthBelowMiddle)) {
					layoutToX = referenceComponentBounds.x + referenceComponentBounds.width + componentLayoutConstraint.dx;
					layoutToY = referenceComponentBounds.y + componentLayoutConstraint.dy
							+ referenceComponentBounds.height - (int) ((2/3)*componentBounds.height);
				}
				else
				if (componentLayoutConstraint.placement.equals(tagLowerLeftCornerIsSixthBelowMiddle)) {
					layoutToX = referenceComponentBounds.x - componentBounds.width - componentLayoutConstraint.dx;
					layoutToY = referenceComponentBounds.y + componentLayoutConstraint.dy
							+ referenceComponentBounds.height - (int) ((2/3)*componentBounds.height);
				}
				else
				if (componentLayoutConstraint.placement.equals(tagUpperRightCornerIsMiddle)) {
					layoutToX = referenceComponentBounds.x + referenceComponentBounds.width + componentLayoutConstraint.dx;
					layoutToY = referenceComponentBounds.y - componentLayoutConstraint.dy - (int) (componentBounds.height/2);
				}
				else
				if (componentLayoutConstraint.placement.equals(tagLowerLeftCorner)) {
					layoutToX = referenceComponentBounds.x - componentBounds.width - componentLayoutConstraint.dx;
					layoutToY = referenceComponentBounds.y + componentLayoutConstraint.dy + referenceComponentBounds.height;
				}
				else
				if (componentLayoutConstraint.placement.equals(tagInsideUpperLeft)) {
					layoutToX = referenceComponentBounds.x + componentLayoutConstraint.dx;
					layoutToY = referenceComponentBounds.y + componentLayoutConstraint.dy;
				}
				else
				if (componentLayoutConstraint.placement.equals(tagInsideUpperRight)) {
					layoutToX = referenceComponentBounds.x + referenceComponentBounds.width - componentBounds.width - componentLayoutConstraint.dx;
					layoutToY = referenceComponentBounds.y + componentLayoutConstraint.dy;
				}
				else
				if (componentLayoutConstraint.placement.equals(tagInsideLowerRight)) {
					layoutToX = referenceComponentBounds.x + referenceComponentBounds.width - componentBounds.width - componentLayoutConstraint.dx;
					layoutToY = referenceComponentBounds.y + referenceComponentBounds.height - componentBounds.height - componentLayoutConstraint.dy;
				}
				else
				if (componentLayoutConstraint.placement.equals(tagInsideLowerLeft)) {
					layoutToX = referenceComponentBounds.x + componentLayoutConstraint.dx;
					layoutToY = referenceComponentBounds.y + referenceComponentBounds.height - componentBounds.height - componentLayoutConstraint.dy;
				}
				else
				if (componentLayoutConstraint.placement.equals(tagAboveLeft)) {
					layoutToX = referenceComponentBounds.x + componentLayoutConstraint.dx;
					layoutToY = referenceComponentBounds.y - componentLayoutConstraint.dy - componentBounds.height;
				}
				else
				if (componentLayoutConstraint.placement.equals(tagUnderLeft) ||
						componentLayoutConstraint.placement.equals(tagBelowLeft)) {
					layoutToX = referenceComponentBounds.x + componentLayoutConstraint.dx;
					layoutToY = referenceComponentBounds.y + componentLayoutConstraint.dy + referenceComponentBounds.height;
				}
				else
				if (componentLayoutConstraint.placement.equals(tagLeftTop)) {
					layoutToX = referenceComponentBounds.x - componentLayoutConstraint.dx - componentBounds.width;
					layoutToY = referenceComponentBounds.y + componentLayoutConstraint.dy;
				}
				else
				if (componentLayoutConstraint.placement.equals(tagRightTop)) {
					layoutToX = referenceComponentBounds.x + componentLayoutConstraint.dx + referenceComponentBounds.width;
					layoutToY = referenceComponentBounds.y + componentLayoutConstraint.dy;
				}
				else
				if (componentLayoutConstraint.placement.equals(tagInsideTop)) {
					layoutToX = referenceComponentBounds.x + referenceComponentBounds.width/2 - componentBounds.width/2 + componentLayoutConstraint.dx;
					layoutToY = referenceComponentBounds.y + componentLayoutConstraint.dy;
				}
				else
				if (componentLayoutConstraint.placement.equals(tagInsideBottom)) {
					layoutToX = referenceComponentBounds.x + referenceComponentBounds.width/2 - componentBounds.width/2 + componentLayoutConstraint.dx;
					layoutToY = referenceComponentBounds.y - componentLayoutConstraint.dy + referenceComponentBounds.height - componentBounds.height;
				}
				else
				if (componentLayoutConstraint.placement.equals(tagInsideLeft)) {
					layoutToX = referenceComponentBounds.x + componentLayoutConstraint.dx;
					layoutToY = referenceComponentBounds.y + referenceComponentBounds.height/2 - componentBounds.height/2 + componentLayoutConstraint.dy;
				}
				else
				if (componentLayoutConstraint.placement.equals(tagInsideRight)) {
					layoutToX = referenceComponentBounds.x - componentLayoutConstraint.dx + referenceComponentBounds.width - componentBounds.width;
					layoutToY = referenceComponentBounds.y + referenceComponentBounds.height/2 - componentBounds.height/2 + componentLayoutConstraint.dy;
				}
				else
				if (componentLayoutConstraint.placement.equals(tagBase)) {
					layoutToX = componentLayoutConstraint.dx; // was 0
					layoutToY = componentLayoutConstraint.dy; // was 0
				}
				else
				// Anchor responses
				if (componentLayoutConstraint.placement.equals(tagRightAnchored)) {
					if (vAnchorOffset != -1) {
					
						layoutToX = referenceComponentBounds.x + componentLayoutConstraint.dx + referenceComponentBounds.width;
					
						MightyRelativeLayout currentCompMRL = GetMRLOfComponent (componentInLayout);
						MightyRelativeLayout refCompMRL = GetMRLOfComponent (componentLayoutConstraint.reference);

						CheckMRLsNotNull (currentCompMRL,refCompMRL);

						layoutToY = refCompMRL.vAnchorOffset+referenceComponentBounds.y-currentCompMRL.vAnchorOffset;
					} else {
						System.out.println("MightyRelativeLayout::layoutContainer -- Error! Trying to right anchor a component but no vertical anchor has been specified.");
						System.exit(1);
					}
				} else 
				if (componentLayoutConstraint.placement.equals(tagLeftAnchored)) {
					if (vAnchorOffset != -1) {
					
						layoutToX = referenceComponentBounds.x - componentLayoutConstraint.dx - componentBounds.width;
					
						MightyRelativeLayout currentCompMRL = GetMRLOfComponent (componentInLayout);
						MightyRelativeLayout refCompMRL = GetMRLOfComponent (componentLayoutConstraint.reference);

						CheckMRLsNotNull (currentCompMRL,refCompMRL);

						layoutToY = refCompMRL.vAnchorOffset+referenceComponentBounds.y-currentCompMRL.vAnchorOffset;
					} else {
						System.out.println("MightyRelativeLayout::layoutContainer -- Error! Trying to left anchor a component but no vertical anchor has been specified.");
						System.exit(1);
					}
				} else 
				if (componentLayoutConstraint.placement.equals(tagTopAnchored)) {
					if (hAnchorOffset != -1) {
						layoutToY = referenceComponentBounds.y - componentLayoutConstraint.dy - componentBounds.height;
						
						MightyRelativeLayout currentCompMRL = GetMRLOfComponent (componentInLayout);
						MightyRelativeLayout refCompMRL = GetMRLOfComponent (componentLayoutConstraint.reference);

						CheckMRLsNotNull (currentCompMRL,refCompMRL);

						layoutToX = refCompMRL.hAnchorOffset+referenceComponentBounds.x-currentCompMRL.hAnchorOffset;
					} else {
						System.out.println("MightyRelativeLayout::layoutContainer -- Error! Trying to top anchor a component but no horizontal anchor has been specified.");
						System.exit(1);
					}
				} else 
				if (componentLayoutConstraint.placement.equals(tagBottomAnchored)) {
					if (hAnchorOffset != -1) {
					
						layoutToY = referenceComponentBounds.y + referenceComponentBounds.height + componentLayoutConstraint.dy - componentBounds.height;
					
						MightyRelativeLayout currentCompMRL = GetMRLOfComponent (componentInLayout);
						MightyRelativeLayout refCompMRL = GetMRLOfComponent (componentLayoutConstraint.reference);

						CheckMRLsNotNull (currentCompMRL,refCompMRL);

						layoutToX = refCompMRL.hAnchorOffset+referenceComponentBounds.x-currentCompMRL.hAnchorOffset;
					} else {
						System.out.println("MightyRelativeLayout::layoutContainer -- Error! Trying to bottom anchor a component but no horizontal anchor has been specified.");
						System.exit(1);
					}
				} else {
					System.out.println("MightyRelativeLayout::layoutContainer -- Major problems, fella, don't know what placement you're specifying.");
					System.exit(1);
				}

				// The layoutToX and layoutToY has been defined, so, reshape it and process after...

				componentInLayout.reshape(layoutToX,layoutToY,componentPreferredSize.width,componentPreferredSize.height);  

				CountForLayoutBoundary(i);
				AdjustIfNecessary(i);
				AnchorProcess(i);
			}

		}
	}


	static MightyRelativeLayout GetMRLOfComponent (Component theComponent) {

		if (theComponent instanceof java.awt.Panel)
			if (((Panel) theComponent).getLayout() instanceof MightyRelativeLayout)
				return ((MightyRelativeLayout) ((Panel) theComponent).getLayout());

		return null;
	}



	static boolean CheckMRLsNotNull (MightyRelativeLayout theCurrentComponent, MightyRelativeLayout theRefComponent) {
		if ((theCurrentComponent == null) || (theRefComponent == null)) {
			System.out.println ("MightyRelativeLayout::CheckMRLsNotNull -- Major problem, dude, placing relative to an anchored component requires both components use MightyRelativeLayout managers.");
			System.exit(1);
		}

		return true;
	}


}


/**
The MightyRelativeConstraint class is used by the 
<a href="MightyRelativeLayout.html">MightyRelativeLayout</a> to store
the information about each component in the MightyRelativeLayout. <p>

<dl>
	<dt> <b>Version:</b>
	<dd> 1.05,  12 January 1996
</dl>
 
<dl>
 	<dt> <b>Known Bugs:</b>
	<dd> None.
</dl>

<dl>
 	<dt> <b>Authors:</b>
 	<dd> <A HREF="http://amber.wpi.edu/~thethe">Benjamin "Quincy" Cabell V</A>,
 	<dd> <A HREF="http://amber.wpi.edu/~thethe/Documents/Besiex/Java/index.html">Besiex Software</A>,
 	<dd> built from Jean-Claude Dufourd's 
			<a href="http://www-elec.enst.fr/java/RelativeLayout.java">RelativeLayout
			Manager</a>
</dl>
*/

class MightyRelativeConstraint {
	Component reference;
	
	int dx, dy;
	
	String placement;
	String name;
	
	String anchor;

	MightyRelativeConstraint(String theName, Component theReference, String thePlacement, int x, int y, String theAnchor) {
		name = theName;
		reference = theReference;
		placement = thePlacement;
		dx = x;
		dy = y;
		anchor = theAnchor;		
	}
}
