package awtExt;
import java.io.IOException ;
import sysExt.* ;


/*
 * @(#)Callback.java	1.0 96/01/05  Sal Cataudella
 *
 * Copyright (c) 1995 Sal Cataudella. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please contact author for other arrangements.
 *
 *
 */

/** 
* Utility class for generating a methodReference to a method which is
* invoked from the awtExt package. The method called back MUST 
* take a CallbackInfo object as a parameter, and have a signature of the form:
* <pre>
* public void calledbackMethod( CallbackInfo cbi );
* </pre>
* The class that the method belongs to must
* be declared public as well.
* @see		sysExt.methodRef
* @author		Sal Cataudella
* @version        1.00     Jan 2 1996
* 
*
*
*/
public abstract class Callback extends Object {
    static String copyright =
	"Copyright C) 1996 by S. Cataudella.  You may use this freely\nprovided that this notice is left here." ;

    /**
    * Creates a GUI callback to a specific method.
    * The called back method *must* have a specific signature, i.e
    * must take a parameter of class CallbackInfo and return void.
    *
    * @param packageString		package of target object
    * @param classString		class of target object
    * @param methodString		string representation of method
    *						to make a reference to  ..
    * @return				reference to method
    * @exception			InvalidMethodRefException  If generation of methodRef fails for any reason  
    */	
    public
    static methodRef newRef(
                            	String packageString,
					String classString,
                            	String methodString )
						throws InvalidMethodRefException {

	return methodRef.newRef( packageString, classString,
					methodString, "awtExt.CallbackInfo");	
    }

   /**
    * newRef() - Convenience method. The target class is extracted
    * from the targetObj, and the target object is then set to targetObj.
    * @param targetObj	package and class string are extracted from
    *				it. The target object is then set to this.
    * @param method	string representation of method to make
    *			       reference to. 
    * @return			reference to method
    * @exception		InvalidMethodRefException If generation of methodRef fails for any reason  
    * @see sysExt.methodRef
    */
    public static methodRef newRef( Object targetObj, String method )
						throws InvalidMethodRefException {

	/*
	* Generate a methodRef to a method which takes an awtExt.CallbackInfo
	* object, and set the target object of the method.
	*/
	try {
		Class aClass = Class.forName("awtExt.CallbackInfo");
		return methodRef.newRef( targetObj,	method, aClass ) ;
	} catch( ClassNotFoundException e) {
		throw new InvalidMethodRefException( "newRef: awtExt.CallbackInfo not found" );
	}
					
    }
 
}

