package awtExt ;

/*
 * @(#)CallbackInfo.java	1.0 96/01/05  Sal Cataudella
 *
 * Copyright (c) 1995 Sal Cataudella. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please contact author for other arrangements.
 *
 *
 */

/**
* Encapsulates the basic data that is passed
* to all called back methods from awtExt objects.
* All called back user defined methods MUST 
* take a CallbackInfo object as a parameter,
* and have a signature of the form:
* <pre>
* public void calledbackMethod( CallbackInfo cbi );
* </pre>
* The class that the method belongs to must
* be declared public as well.
* @see		sysExt.methodRef
* @author		Sal Cataudella
* @version        1.0     Jan 05 1996
***/
public class CallbackInfo {
	/**
	* The event that just occurred.
	* @see java.awt.Event
	**/
	public java.awt.Event evt ;	
	
	/**
	* The standard 'what' object passed to awt components.
	* For example, in the case of a button this is the label.
	* Please refer to std java doc for further details.
	**/
	public Object what ;
	 
	/**
	* The current awt object...
	* not null only in Choice objects.
	**/
	public Object thisAwtObj ;

	/**
	* Indicates whether the event has been processed by this 
	* awtExt object's callback.
	* Default is false, which causes the event to propagate to
	* this object's container.
	* Setting this true will stop the event propagation.
	**/ 
	public boolean processed = false ;

	/**
	* Constructor. Used to initialize the CallbackInfo object
	* @param	e awt event that just occurred.
	* @param	w awtExt component dependent parameter. This is 
	*		  usually the label of the GUI object. Refer to
	*		  the specific awtExt object for more details.
	* @param	tao Points to the current awtExt object. Used for
	* accessing awtExt specific functionality.
	* @see java.awt.Event
	*
	**/
       public CallbackInfo( java.awt.Event e, Object w, Object tao ) { 
		evt = e ; what = w; thisAwtObj = tao ;
	} ;

	/**
	* Alternate constructor.
	* @param	e awt event that just occurred.
	* @param	w awtExt component dependent parameter. This is 
	*		  usually the label of the GUI object. Refer to
	*		  the specific awtExt object for more details.
	**/
       public CallbackInfo( java.awt.Event e, Object w) { 
		evt = e ; what = w; thisAwtObj = null ;
	} ;

};
