package awtExt ;

/*
 * @(#)Checkbox.java	1.0 96/01/05  Sal Cataudella
 *
 * Copyright (c) 1995 Sal Cataudella. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please contact author for other arrangements.
 *
 *
 */



import sysExt.* ;


/**
* Extension of standard awt Checkbox object.  Provides a callback
* mechanism for invoking a user defined method, when an
* ACTION_EVENT occurs within this awtExt component. The called back
* method and target object must have a specific signature.
* This allows you to separate your application functionality
* from the GUI. i.e, no need to subclass the Checkbox object in order
* to invoke your user defined code.
* @see		awtExt.Callback
* @see		java.awt.Checkbox
* @author		Sal Cataudella
* @version        1.0     Jan 05 1996
**/
public class Checkbox extends java.awt.Checkbox {
	methodRef actionMethodRef = null ;
	

	/**
	* Default Constructor. 
	**/
	public Checkbox() {
		super() ;
	}


	/**
	* Constructor. Takes a string as a parameter, to be
	* used as the label.
	* @param aLabel label for the checkbox
	**/
	public Checkbox( java.lang.String aLabel  ) {
		super( aLabel ) ;
	}

	/**
	* Constructor. 
	* @param aLabel label for the checkbox
	* @param group Checkbox group
	* @param state checkbox' state
	**/
	public Checkbox( java.lang.String aLabel,
				java.awt.CheckboxGroup group, boolean state  ) {
		super( aLabel, group, state ) ;
	}


	/**
	* Constructor.
	* @param aLabel label for the checkbox.
	* @param aMethodRef  Fully resolved methodRef, i.e must have
	* the target object set.
	* @see sysExt.methodRef#setTarget
	**/
	public Checkbox( java.lang.String aLabel, methodRef mr  ) {
		super( aLabel ) ;
		actionMethodRef = mr ;
	}	

	/**
	* Constructor.
	* @param aLabel label for the checkbox
	* @param target object which is target of the called back method
	* @param method string representation of method to invoke on
	* target object.
	* @exception InvalidMethodRefException If generation of callback fails
	**/
	public Checkbox( java.lang.String aLabel,
				    Object target, String method  ) throws InvalidMethodRefException {
		super( aLabel ) ;
		actionMethodRef = Callback.newRef( target, method );
	}

	/**
	* Adds user defined methodRef. This is invoked when a
	* ACTION_EVENT occurs. The methodRef must be
	* defined using the utility class awtExt.Callback
	* @param aMethodRef ACTION_EVENT event handling method.
	* @see awtExt.Callback
	* @see sysExt.methodRef
	* 
	**/
	public void	addActionCallback( methodRef aMethodRef ) {
		actionMethodRef = aMethodRef ;
	}

	/**
	* Called when an action occurs. This in turn calls back
	* a user defined method on a target object. The called back
	* method *AND* the target object must have a specific signature.
	* Since the action event for this checkbox is *always* handled 
	* here, this action() method returns false to stop the event
	* propagation.
	* @param evt The action event that just occurred.
	* @param obj The label of this object.
	* @see 	awtExt.Callback
	* @see	awtExt.CallbackInfo
	**/
	public final boolean action( java.awt.Event evt,
					Object obj ) {
		CallbackInfo cbInfo = new CallbackInfo ( evt, obj, this ) ;
		if(actionMethodRef != null )
			actionMethodRef.doAction( cbInfo ) ; 
		return cbInfo.processed;
	}
} ;

