package awtExt ;

/*
 * @(#)Choice.java	1.0 96/01/05  Sal Cataudella
 *
 * Copyright (c) 1995 Sal Cataudella. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please contact author for other arrangements.
 *
 *
 */


import sysExt.* ;


/**
* Extension of standard awt Choice object. Provides a callback
* mechanism for invoking a user defined method, when an
* ACTION_EVENT occurs within this awtExt component. The called
* back method and target object must have a specific signature.
* This allows you to separate your application functionality
* from the GUI. i.e, no need to subclass the Choice object in order
* to invoke your user defined code.
* <pre>
* </pre>
* NOTE: This component does NOT  use the general EventDispatch
* mechanism, as the only event of interest is the ACTION_EVENT.
* @see		awtExt.Callback
* @see		java.awt.Choice
* @author		Sal Cataudella
* @version        1.0     Jan 05 1996
**/
public class Choice extends java.awt.Choice {
	methodRef actionMethodRef = null ;
	

	/**
	* Default Constructor. 
	**/
	public Choice() {
		super() ;
	}

	/**
	* Constructor.
	* @param mr method reference to callbackable method.
	* mr must have the target object set.
	* @see sysExt.methodRef#setTarget
	**/
	public Choice(methodRef mr  ) {
		super() ;
		actionMethodRef = mr ;
	}	



	/**
	* Constructor. An exception may be thrown because it generates
	* a methodRef from the target, method parameter pair.
	* @param target object which is target of the called back method
	* @param method string representation of method to invoke on
	* target object.
	* @exception InvalidMethodRefException  generated if creation of callback fails
	**/
	public Choice( Object target, String method  ) throws InvalidMethodRefException  {
		actionMethodRef = Callback.newRef( target, method ) ;
	}

	/**
	* Adds user defined methodRef. This is invoked when a
	* ACTION_EVENT occurs. The methodRef must be
	* defined using the utility class awtExt.Callback
	* @param aMethodRef ACTION_EVENT event handling method.
	* aMethodRefmust have the target object set.
	* @see sysExt.methodRef#setTarget
	* @see awtExt.Callback
	* @see sysExt.methodRef
	* 
	**/
	public void	addActionCallback( methodRef aMethodRef ) {
		actionMethodRef = aMethodRef ;
	}

	/**
	* Called when an action occurs. This in turn calls back
	* a user defined method on a target object. The called back
	* method *AND* the target object must have a specific signature.
	* @param evt The action event that just occurred.
	* @param obj The label of this object.
	* @see 	awtExt.Callback
	* @see	awtExt.CallbackInfo
	**/
	public boolean action( java.awt.Event evt,
					Object obj ) {
		CallbackInfo cbInfo = new CallbackInfo ( evt, obj, this ) ;
		actionMethodRef.doAction( (Object)cbInfo ) ; 
		return cbInfo.processed;
	}
} ;

