package awtExt ;

/*
 * @(#)EventDispatch.java	1.0 96/01/05  Sal Cataudella
 *
 * Copyright (c) 1995 Sal Cataudella. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please contact author for other arrangements.
 *
 *
 */


import sysExt.*;
import java.awt.Event ;


/**
* Dispatches an event to a user defined method when that event occurs.
* To attach a callbackable method to an event, simply assign a methodRef
* object to the variable with the same name as the event.
* For a given awtExt GUI object, provides an easily configurable event handling
* method dispatcher.
* For example, to process KEY_PRESS events in an awtExt.TextField object:
* <pre>
* 	awtExt.TextField aField = new awtExt.TextField(...);
*	aField.eventDispatch.KEY_PRESS = Callback.newRef( myObj, "myMethod" );
* </pre>
* When a KEY_PRESS event occurs, <pre> myObj.myMethod( cbi ) </pre>
* will be dispatched. <pre> myMethod() </pre> can then control
* the event's propagation as well as modify its state.
* <pre>
* </pre>
* Note: a default event handler is installed for the WINDOW_DESTROY
* event. This will cause the Java application to exit. If it's an applet,
* the window will be hidden and disposed. To deactivate it, set the
* class variable USE_DEFAULT_WINDOW_DESTROY to false.
* @see		java.awt.Event
* @see		awtExt.Callback
* @see		awtExt.CallbackInfo#processed
* @see		awtExt.CallbackInfo
* @see		sysExt.methodRef
* @see		awtExt.EventDispatch#USE_DEFAULT_WINDOW_DESTROY 
* @author		Sal Cataudella
* @version        1.0     Jan 05 1996
**/
public final class EventDispatch extends Object {
//
// The callback hooks.....
//
	/**
	* User installed methodRef for an action event
	* @see java.awt.Event#ACTION_EVENT
	**/
  public methodRef ACTION_EVENT= null;
	/**
	* User installed methodRef for a got focus event
	* @see java.awt.Event#GOT_FOCUS
	**/
  public methodRef GOT_FOCUS= null;
	/**
	* User installed methodRef for a key action event
	* @see java.awt.Event#KEY_ACTION
	**/
  public methodRef KEY_ACTION= null;
	/**
	* User installed methodRef for a key action release event
	* @see java.awt.Event#KEY_ACTION_RELEASE
	**/
  public methodRef KEY_ACTION_RELEASE= null;
	/**
	* User installed methodRef for a key press event
	* @see java.awt.Event#KEY_PRESS
	**/	
  public methodRef	KEY_PRESS= null;
	/**
	* User installed methodRef for a key release event
	* @see java.awt.Event#KEY_RELEASE
	**/
  public methodRef	KEY_RELEASE= null;
	/**
	* User installed methodRef for a list deselect event
	* @see java.awt.Event#LIST_DESELECT
	**/
  public methodRef	LIST_DESELECT= null;
	/**
	* User installed methodRef for a list select event
	* @see java.awt.Event#LIST_DESELECT
	**/
  public methodRef	LIST_SELECT= null;
	/**
	* User installed methodRef for a load file event
	* @see java.awt.Event#LOAD_FILE
	**/
  public methodRef	LOAD_FILE= null;
	/**
	* User installed methodRef for a lost focus event
	* @see java.awt.Event#LOST_FOCUS
	**/
  public methodRef	LOST_FOCUS= null;
	/**
	* User installed methodRef for a mouse down event
	* @see java.awt.Event#MOUSE_DOWN
	**/
  public methodRef	MOUSE_DOWN= null;
	/**
	* User installed methodRef for a mouse drag event
	* @see java.awt.Event#MOUSE_DRAG
	**/	
  public methodRef	MOUSE_DRAG= null;
	/**
	* User installed methodRef for a mouse enter event
	* @see java.awt.Event#MOUSE_ENTER
	**/
  public methodRef	MOUSE_ENTER= null;
	/**
	* User installed methodRef for a mouse exit event
	* @see java.awt.Event#MOUSE_EXIT
	**/
  public methodRef	MOUSE_EXIT= null;
	/**
	* User installed methodRef for a mouse move event
	* @see java.awt.Event#MOUSE_MOVE
	**/
  public methodRef	MOUSE_MOVE= null;
	/**
	* User installed methodRef for a mouse up event
	* @see java.awt.Event#MOUSE_UP
	**/
  public methodRef	MOUSE_UP= null;
	/**
	* User installed methodRef for a save file event
	* @see java.awt.Event#SAVE_FILE
	**/
  public methodRef	SAVE_FILE= null;
	/**
	* User installed methodRef for an scroll absolute event
	* @see java.awt.Event#SCROLL_ABSOLUTE
	**/
  public methodRef	SCROLL_ABSOLUTE= null;
	/**
	* User installed methodRef for a scroll line down event
	* @see java.awt.Event#SCROLL_LINE_DOWN
	**/
  public methodRef	SCROLL_LINE_DOWN= null;
	/**
	* User installed methodRef for a scroll line up event
	* @see java.awt.Event#SCROLL_LINE_UP
	**/
  public methodRef	SCROLL_LINE_UP= null;
	/**
	* User installed methodRef for a scroll page down event
	* @see java.awt.Event#SCROLL_PAGE_DOWN
	**/
  public methodRef	SCROLL_PAGE_DOWN= null;
 	/**
	* User installed methodRef for a scroll page down event
	* @see java.awt.Event#SCROLL_PAGE_UP
	**/
  public methodRef	SCROLL_PAGE_UP= null;
	/**
	* User installed methodRef for a window destroy event
	* @see java.awt.Event#WINDOW_DESTROY
	**/
  public methodRef	WINDOW_DESTROY= null;
	/**
	* User installed methodRef for a window iconify event
	* @see java.awt.Event#WINDOW_ICONIFY
	**/
  public methodRef	WINDOW_ICONIFY= null;
	/**
	* User installed methodRef for a window deiconify event
	* @see java.awt.Event#WINDOW_DEICONIFY
	**/
  public methodRef	WINDOW_DEICONIFY= null;
	/**
	* User installed methodRef for a window expose event
	* @see java.awt.Event#WINDOW_EXPOSE
	**/
  public methodRef	WINDOW_EXPOSE= null;
	/**
	* User installed methodRef for a window moved event
	* @see java.awt.Event#WINDOW_MOVED
	**/
  public methodRef	WINDOW_MOVED= null ;

	/**
	* Container of this EventDispatch object. Passed to each
	* user defined event handling method via the CallbackInfo parameter
	* @see CallbackInfo#thisAwtObj 
	**/
  public Object container = null ;

    /**
    * Constructor. Attempts to install a default WINDOW_DESTROY 
    * event handler method. This behaviour can be turned off by
    * setting the USE_DEFAULT_WINDOW_DESTROY class variable to false.
    * @see #USE_DEFAULT_WINDOW_DESTROY 
    */
    public EventDispatch( Object aContainer ) {
	container = aContainer ;
	if( !USE_DEFAULT_WINDOW_DESTROY ) 
		return ;

	try {
		WINDOW_DESTROY = methodRef.newRef( "awtExt", "EventDispatch",
								   "defaultWindowDestroy",
								   "awtExt.CallbackInfo");
		WINDOW_DESTROY.setTarget(this); 
	} catch (Exception e ) {
		System.out.println( "awtExt.EventDispatch exception") ;
		e.printStackTrace();
	} ;
	
	/*
	** Test to determine if we are executing within an applet..
	** Do this test only once. Basically I attempt to generate
	** a security exception on purpose, to determine if within 
	** an applet or not.
	** There *has* to be a better way...
	*/
	if( -1 == inApplet ) { //First time this is tried?
		try {
			SecurityManager s = new AppletTester();
			inApplet = 0 ; // If here then it's an application	
		} catch( SecurityException e ) {
			inApplet = 1 ; // Exception caught. Must be an applet
		}
	}

    }	


    /**
    * Dispatches an event to a user defined event handling method. For a
    * given event, checks to see if the corresponding event handling
    * methodRef is set and if so, dispatches the event to it. 
    * @param evt The current event.
    * @return event processed status.
    * @see awtExt.CallbackInfo
    * @see awt.Event 
    */
    public final boolean handle( Event evt ) {
	CallbackInfo cbi = new CallbackInfo( evt, this, container ) ;

	switch( evt.id ) {
	case Event.ACTION_EVENT:
		if(ACTION_EVENT != null )
			ACTION_EVENT.doAction( cbi ) ;
		break;
	case Event.GOT_FOCUS:
		if(GOT_FOCUS!= null )
			GOT_FOCUS.doAction( cbi ) ; 
		break ;
	case Event.KEY_ACTION:
		if(KEY_ACTION!= null )
			KEY_ACTION.doAction( cbi ) ; 
		break ;

	case Event.KEY_ACTION_RELEASE:
		if(KEY_ACTION_RELEASE!= null )
			KEY_ACTION_RELEASE.doAction( cbi ) ; 
		break ;
	case Event.KEY_PRESS:
		if(KEY_PRESS!= null )
			KEY_PRESS.doAction( cbi ) ; 
		break ;
	case Event.KEY_RELEASE:
		if(KEY_RELEASE!= null )
			KEY_RELEASE.doAction( cbi ) ; 
		break ;

	case Event.LIST_DESELECT:
		if(LIST_DESELECT!= null )
			LIST_DESELECT.doAction( cbi ) ; 
		break ;

	case Event.LIST_SELECT:
		if(LIST_SELECT!= null )
			LIST_SELECT.doAction( cbi ) ; 
		break ;

	case Event.LOAD_FILE:
		if(LOAD_FILE!= null )
			LOAD_FILE.doAction( cbi ) ; 
		break ;

	case Event.LOST_FOCUS:
		if(LOST_FOCUS!= null )
			LOST_FOCUS.doAction( cbi ) ; 
		break ;

	case Event.MOUSE_UP:
		if(MOUSE_UP!= null )
			MOUSE_UP.doAction( cbi ) ; 
		break ;

	case Event.MOUSE_DOWN:
		if(MOUSE_DOWN!= null )
			MOUSE_DOWN.doAction( cbi ) ; 
		break ;

	case Event.MOUSE_DRAG:
		if(MOUSE_DRAG!= null )
			MOUSE_DRAG.doAction( cbi ) ; 
		break ;

	case Event.MOUSE_ENTER:
		if(MOUSE_ENTER!= null )
			MOUSE_ENTER.doAction( cbi ) ; 
		break ;

	case Event.MOUSE_EXIT:
		if(MOUSE_EXIT!= null )
			MOUSE_EXIT.doAction( cbi ) ; 
		break ;

	case Event.MOUSE_MOVE:
		if(MOUSE_MOVE!= null )
			MOUSE_MOVE.doAction( cbi ) ; 
		break ;

	case Event.SAVE_FILE:
		if(SAVE_FILE!= null )
			SAVE_FILE.doAction( cbi ) ; 
		break ;

	case Event.SCROLL_ABSOLUTE:
		if(SCROLL_ABSOLUTE!= null )
			SCROLL_ABSOLUTE.doAction( cbi ) ; 
		break ;

	case Event.SCROLL_LINE_DOWN:
		if(SCROLL_LINE_DOWN!= null )
			SCROLL_LINE_DOWN.doAction( cbi ) ; 
		break ;

	case Event.SCROLL_LINE_UP:
		if(SCROLL_LINE_UP!= null )
			SCROLL_LINE_UP.doAction( cbi ) ; 
		break ;

	case Event.SCROLL_PAGE_DOWN:
		if(SCROLL_PAGE_DOWN!= null )
			SCROLL_PAGE_DOWN.doAction( cbi ) ; 
		break ;

	case Event.SCROLL_PAGE_UP:
		if(SCROLL_PAGE_UP!= null )
			SCROLL_PAGE_UP.doAction( cbi ) ; 
		break ;

	case Event.WINDOW_ICONIFY:
		if(WINDOW_ICONIFY!= null )
			WINDOW_ICONIFY.doAction( cbi ) ;
		break;

	case Event.WINDOW_DEICONIFY:
		if(WINDOW_DEICONIFY != null )
			WINDOW_DEICONIFY.doAction( cbi ) ;
		break;

	case Event.WINDOW_EXPOSE:
		if(WINDOW_EXPOSE!= null )
			WINDOW_EXPOSE.doAction( cbi ) ; 
		break ;

	case Event.WINDOW_MOVED:
		if(WINDOW_MOVED!= null )
			WINDOW_MOVED.doAction( cbi ) ; 
		break ;

	case Event.WINDOW_DESTROY:
		if(WINDOW_DESTROY!= null )
			WINDOW_DESTROY.doAction( cbi ) ; 
		break ;
	}
	return cbi.processed;
    }


    /**
    * The default window destroy callback method.
    * This handler method is installed as the default,
    * if the USE_DEFAULT_WINDOW_DESTROY class variable is set to
    * true. Its behaviour is to hide and dispose of the frame.
    * In addition, if it's an application it will perform a
    * System.exit(0)
    */
    public void defaultWindowDestroy( CallbackInfo cbi ) {
	System.out.println("In defaultWindowDestroy..." );
	if( cbi.evt.target instanceof java.awt.Frame ) {
		java.awt.Frame aWindow = (java.awt.Frame)cbi.evt.target ;
		aWindow.hide() ;
		aWindow.dispose() ;
		if( inApplet == 0 )
			System.exit(0);
		cbi.processed = true ;
	} else
		System.out.println( "target is a " + cbi.evt.target.getClass().getName() ) ;
    }



    /**
    * Controls attachment of default window destroy event handler
    * Set to true by default. Set this to false if you don't
    * want a default window destroy handler
    */
    public static boolean USE_DEFAULT_WINDOW_DESTROY = true ;


    static int inApplet = -1 ;


}

class AppletTester extends SecurityManager {
	AppletTester() {
		super() ;
	}
}

