package awtExt ;


/*
 * @(#)List.java	1.0 96/01/05  Sal Cataudella
 *
 * Copyright (c) 1995 Sal Cataudella. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please contact author for other arrangements.
 *
 *
 */

import sysExt.* ;



/**
* Extension of standard awt List. Provides a callback
* mechanism for invoking user defined event handling methods.
* A callback method can be attached on each and every event 'hook'
* in the List's eventDispatch object. 
* @see		awtExt.EventDispatch
* @see		awtExt.Callback
* @see		java.awt.Event
* @see		java.awt.List
* @author		Sal Cataudella
* @version        1.0     Jan 05 1996
**/
public class List extends java.awt.List {
	/**
	* Event dispatcher. Dispatches the event to the appropriate
	* user defined event handling method.
	* @see	awt.Event
	* @see	awtExt.EventDispatch
	* @see	awtExt.Callback
	* 
	**/
	public EventDispatch eventDispatch = new EventDispatch( this ) ;

	/**
	* Default Constructor. 
	**/
	public List() {
		super() ;
	}

	/**
	* Constructor, creates a list with numRows rows and a single
	* or multi-selection mechanism.
	* @param numRows Number of rows for this list
	* @param mSel true if to use multiple selection
	**/
	public List( int numRows, boolean mSel) {
		super( numRows, mSel) ;
	}


	/**
	* event handler. Dispatches the event to the appropriate
	* attached methodRef event handler . User defined method(s) can control
	* the event propagation by setting the 'processed' instance variable
	* in the CallbackInfo object.
	* @param	evt The event.
	* @see	awt.Event
	* @see	awtExt.EventDispatch
	* @see	awtExt.CallbackInfo#processed
	*
	**/
      public final boolean handleEvent(java.awt.Event evt) {
		return eventDispatch.handle( evt ) ;
      }
} ;

