package awtExt ;



/*
 * @(#)MenuBar.java	1.0 96/01/05  Sal Cataudella
 *
 * Copyright (c) 1995 Sal Cataudella. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please contact author for other arrangements.
 *
 *
 */


import sysExt.* ;



/**
* Extension of standard awt MenuBar. Provides a callback
* mechanism for invoking a user defined method, when an
* ACTION_EVENT occurs within this awtExt component. The called
* back method and target object must have a specific signature.
* This allows you to separate your application functionality
* from the GUI. i.e, no need to subclass the MenuBar in order
* to invoke your user defined code.
* <pre>
* </pre>
* NOTE: This component does NOT  use the general EventDispatch
* mechanism, as the only event of interest is the ACTION_EVENT.
* @see		awtExt.Callback
* @see		java.awt.MenuBar
* @author		Sal Cataudella
* @version        1.0     Jan 05 1996
**/
public class MenuBar extends java.awt.MenuBar {
	methodRef eventHandlerMethodRef= null ;
	


	/**
	* Constructor. 
	**/
	public MenuBar() {
		super() ;
	}

	/**
	* Constructor.
	* @param mr method reference to callbackable method.
	* mr must have the target object set.
	* @see sysExt.methodRef#setTarget
	**/
	public MenuBar(methodRef mr  ) {
		super() ;
		eventHandlerMethodRef= mr ;
	}	

	/**
	* Constructor.
	* @param target object which is target of the called back method
	* @param method string representation of method to invoke on
	* target object.
	* @exception InvalidMethodRefException If generation of callback fails for any reason.
	**/
	public MenuBar(Object target, String method  ) throws InvalidMethodRefException {
		super() ;
		eventHandlerMethodRef= Callback.newRef( target, method ) ;
	}

	/**
	* Adds a user defined methodRef. This is invoked when a
	* ACTION_EVENT occurs. The methodRef must be
	* defined using the utility class awtExt.Callback
	* @param aMethodRef  Fully resolved methodRef, i.e must have
	* the target object set.
	* @see awtExt.Callback
	* @see sysExt.methodRef
	* 
	**/
	public void	addActionCallback( methodRef aMethodRef ) {
		eventHandlerMethodRef= aMethodRef ;
	}

	/**	
	* Posts the specified event to the menubar.
	* The event will propagate to its  frame container,
	* if it's not handled in here, or if the CallbackInfo.processed
	* is false.
 	* @param evt the event which is to take place
	* @see  awtExt.Callback
	*/	
	public final boolean postEvent(java.awt.Event evt) {

		if( handleEvent(evt) )
			return true;
 		else  return (this.getParent()).postEvent(evt) ;	
	}


 	/**
	* Handles the event. Invokes user defined method, if there
	* is one. Returns CallbackInfo.processed to control
	* event propagation
 	* @param evt the event which is to take place
	* @see  awtExt.CallbackInfo
	* @see  awtExt.CallbackInfo#processed
	**/
	public final boolean handleEvent( java.awt.Event evt) {
		CallbackInfo cbInfo = new CallbackInfo ( evt, this ) ;
		if(eventHandlerMethodRef != null )
			eventHandlerMethodRef.doAction( cbInfo ) ; 
		return cbInfo.processed;
    }
} ;

