package awtExt ;



/*
 * @(#)Scrollbar.java	1.0 96/01/05  Sal Cataudella
 *
 * Copyright (c) 1995 Sal Cataudella. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please contact author for other arrangements.
 *
 *
 */


import sysExt.* ;


/**
* Extension of standard awt Scrollbar. Provides a callback
* mechanism for invoking a user defined method, when a
* SCROLL_* event  occurs within this awtExt component. The called
* back method and target object must have a specific signature.
* This allows you to separate your application functionality
* from the GUI. i.e, no need to subclass the Scrollbar in order
* to invoke your user defined code.
* @see		awtExt.EventDispatch
* @see		awtExt.Callback
* @see		java.awt.Scrollbar
* @author		Sal Cataudella
* @version        1.0     Jan 05 1996
**/


public class Scrollbar extends java.awt.Scrollbar {
	/**
	* Event dispatcher. Dispatches the SCROLL_* event to the appropriate
	* user defined method.
	* @see	java.awt.Event
	* @see	awtExt.EventDispatch
	* @see	awtExt.Callback
	* 
	**/
	public EventDispatch eventDispatch = new EventDispatch( this ) ;


	/**
	* Default constructor. 
	**/
	public Scrollbar() {
		super() ;
	}

	/**
	* Constructs a new scrollbar with a specific orientation
	* @param orient scrollbar's orientation
	* @exception IllegalArgumentException if invalid argument
	**/
	public Scrollbar( int orient ) throws IllegalArgumentException {
		super( orient ) ;
	}	

	/**
	* Constructs a new scrollbar with the specific values
	* @param orient orientation
	* @param value  current value
	* @param visible size of scrollable region
	* @param min min value of scrollbar
	* @param max max value of scrollbar
	**/
	public Scrollbar( int orient,
				int value,
				int visible,
				int min,
				int max) {
		super( orient, value, visible, min, max ) ;
	}


	/**
	* Constructs a new scrollbar and sets the SCROLL_* event handlers.
	* @param mr method reference to callbackable method.
	* mr must have the target object set.
	* @see sysExt.methodRef#setTarget
	**/
	public Scrollbar(methodRef mr  ) {
		super() ;
		initScrollbarEventHandlers( mr ) ;
	}

	/**
	* Constructs a new scrollbar with a specific orientation
	* and sets the SCROLL_* events handlers.
	* @param orient scrollbar's orientation
	* @param mr method reference to callbackable method.
	* mr must have the target object set.
	* @exception IllegalArgumentException if invalid argument
	* @see sysExt.methodRef#setTarget
	**/
	public Scrollbar( int orient, methodRef mr  )
				throws IllegalArgumentException {
		super( orient ) ;
		initScrollbarEventHandlers( mr );
	}	

	/**
	* Constructs a new scrollbar with the specific values
	* and sets the SCROLL_* events handlers.
	* @param orient orientation
	* @param value  current value
	* @param visible size of scrollable region
	* @param min min value of scrollbar
	* @param max max value of scrollbar
	* @param mr method reference to callbackable method.
	* mr must have the target object set.
	**/
	public Scrollbar( int orient,
				int value,
				int visible,
				int min,
				int max,
				methodRef mr  ) {
		super( orient, value, visible, min, max ) ;
		initScrollbarEventHandlers( mr );
	}


	
	/**
	* Adds user defined methodRef. This is invoked when a
	* SCROLL_* event occurs. The methodRef must be
	* defined using the utility class awtExt.Callback
	* @param aMethodRef SCROLL_* event handling method.
	* @see awtExt.Callback
	* @see sysExt.methodRef
	* 
	**/
	public void	addCallback( methodRef aMethodRef ) {
		initScrollbarEventHandlers( aMethodRef );
	}



	/**
	* event handler. Dispatches the event to the appropriate
	* attached methodRef. User defined method(s) can control
	* the event propagation by setting the processed instance variable
	* in the CallbackInfo object.
	* @param	evt The event.
	* @see	awt.Event
	* @see	awtExt.EventDispatch
	* @see	awtExt.CallbackInfo#processed
	*
	**/
      public final boolean handleEvent(java.awt.Event evt) {
		return eventDispatch.handle( evt ) ;
      }


	/**
	* Utility function.
	**/
	void initScrollbarEventHandlers( methodRef aMethodRef ) {
		eventDispatch.SCROLL_LINE_UP   = 
		eventDispatch.SCROLL_LINE_DOWN = 
		eventDispatch.SCROLL_PAGE_UP   = 
		eventDispatch.SCROLL_PAGE_DOWN = 
		eventDispatch.SCROLL_ABSOLUTE  = aMethodRef ;
	}

} ;

