package sysExt ;
/*
 * @(#)ClassBuilder.java	1.0 96/01/05  Sal Cataudella
 *
 * Copyright (c) 1995 Sal Cataudella. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please contact author for other arrangements.
 *
 *
 */
import java.io.IOException;
import java.io.DataOutputStream;
import java.io.FileOutputStream;
import java.io.FileInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.util.Hashtable ;


class ClassBuilder extends ClassLoader implements RuntimeConstants {
    static boolean debug = false ;

    static Hashtable cache = new Hashtable() ;

    static final short constantPoolCount = 30 ;
    static int staticCPEntries = 0 ;
    static ConstAttrInfo lst[]= new ConstAttrInfo[ constantPoolCount ];
    static ConstAttrInfo CP[] = new ConstAttrInfo [ constantPoolCount ] ;

    //
    // The various constants that are to be placed in the
    // constant pool...
    // These are statics because they need to be
    // initialized only once, and don't depend on any
    // parameters passed to methodRef...
    //
    static	ConstStringInfo doActionSig ;
    static	ConstStringInfo constantValue ;
    static	ConstStringInfo O ;
    static	ConstStringInfo Exceptions ;
    static	ConstStringInfo LineNumberTable;
    static	ConstStringInfo SourceFile;
    static	ConstStringInfo baseClassStr;
    static	ConstStringInfo LocalVariables ;
    static	ConstStringInfo Code;
    static	ConstStringInfo doAction ;
    static	ConstStringInfo _init_;
    static	ConstStringInfo Ljava_lang_Object;
    static	ConstStringInfo voidSig ;

    static	ConstNameAndTypeInfo targetObjNAT ;
    static	ConstClassInfo superClass;
    static	ConstFieldInfo targetObj;
    static  ConstNameAndTypeInfo initNAT ;
    static  ConstMethodInfo initMethodConst;



    static {
	//
	// Init. the  constants that are to be placed in the
 	// constant pool...
 	//
	int i = 0 ;
    	lst[i++]=doActionSig = new ConstStringInfo("(Ljava/lang/Object;)V") ;
	lst[i++]=constantValue = new ConstStringInfo("ConstantValue") ;
	lst[i++]=O = new ConstStringInfo( "O"  ) ;
	lst[i++]=Exceptions = new ConstStringInfo("Exceptions") ;
	lst[i++]=LineNumberTable = new ConstStringInfo("LineNumberTable") ;
	lst[i++]=SourceFile = new ConstStringInfo("SourceFile" ) ;
	lst[i++]=baseClassStr = new ConstStringInfo("sysExt/methodRef" ) ;
	lst[i++]=LocalVariables = new ConstStringInfo("LocalVariables") ;
	lst[i++]=Code = new ConstStringInfo("Code") ;
	lst[i++]=doAction = new ConstStringInfo("doAction") ;	
	lst[i++]=_init_ = new ConstStringInfo("<init>") ;
	lst[i++]=Ljava_lang_Object = new ConstStringInfo("Ljava/lang/Object;");
	lst[i++]=voidSig  = new ConstStringInfo("()V");	
	lst[i++]=targetObjNAT = new ConstNameAndTypeInfo(O, Ljava_lang_Object);
	lst[i++]=superClass = new ConstClassInfo(baseClassStr) ;
	lst[i++]=targetObj = new ConstFieldInfo(superClass, targetObjNAT) ;	
	lst[i++]=initNAT = new ConstNameAndTypeInfo(_init_, voidSig ) ;	
	lst[i++]=initMethodConst = new ConstMethodInfo(superClass, initNAT) ;

	staticCPEntries = i ;
	//
 	// This entry in const pool should never be used ...
	//
	CP[0] = null ;
		
    }

    String functorClassSignature ;

    //
    // COnstants in the constant pool which
    // depend on parameters passed to methodRef.newRef()...
    //
    ConstStringInfo action ;		// string representation of action method
    ConstStringInfo javaFile ;	// functor (pseudo)source file 
    ConstStringInfo thisClassStr;	// current class string
    ConstStringInfo targetClassStr; // target object's class string
    ConstClassInfo thisClass;		// current class
    ConstClassInfo targetClass;	// target object's class
    ConstNameAndTypeInfo doActionNAT;//name and type of action method
    ConstMethodInfo actionMethodConst;//action method...

    ConstStringInfo paramClassString;
    ConstStringInfo targetMethodSig;
    ConstClassInfo  paramClass;

    boolean paramClassIsObject = true ;

    //----------------------------------------------
    // buildClassFile() - Generates a class file.
    // should be thread safe...
    //----------------------------------------------
    synchronized	Class buildClassFile(	String packageString,
							String classString,
							String methodString,
							String functorClassSignature,
							String functorClassName,
							String aParamClassString,
							boolean overrideProtectStatus  
                                       )	throws	Exception {


	ByteArrayOutputStream f = new ByteArrayOutputStream () ;		
	DataOutputStream out = new DataOutputStream( f ) ;


	this.functorClassSignature = functorClassSignature ;

	int i = staticCPEntries ;

	String paramType = null ;
	if( 0 != aParamClassString.compareTo( "java.lang.Object" ) ) {
		paramClassIsObject = false ;
		aParamClassString = aParamClassString.replace( '.', '/' ) ;
		paramType = new String( aParamClassString );
	} else paramClassIsObject = true ;
	
	//
	// Create the various constant strings needed for a std java class...
	//

	lst[i++] = action = new ConstStringInfo(methodString) ;
	lst[i++] = javaFile = new ConstStringInfo(functorClassName + ".java");
	lst[i++] = thisClassStr = new ConstStringInfo("methodRefClasses/" +
								functorClassName );

	lst[i++] =	targetClassStr = new ConstStringInfo(
								packageString != null ? 
								packageString + "/" + classString :
								classString) ;
	lst[i++] = thisClass = new ConstClassInfo(thisClassStr) ;
	lst[i++] = targetClass = new ConstClassInfo(targetClassStr) ;

	if( !paramClassIsObject ) {
		lst[i++] = paramClassString = new ConstStringInfo( aParamClassString ) ;
		lst[i++] = targetMethodSig = new ConstStringInfo( "(L" + paramType + ";)V") ;	
		lst[i++] = paramClass = new ConstClassInfo( paramClassString ) ;
	}


	lst[i++] = doActionNAT = new ConstNameAndTypeInfo(action,
							paramClassIsObject ?
							doActionSig:
							targetMethodSig);

	lst[i++] = actionMethodConst = new ConstMethodInfo(targetClass,
							doActionNAT);

	//
	// Now transfer all constants from the temp lst to
	// the 'real' CP array. Their ordering in the list depends on the
	// constants' 'level' value ...

	int numCPEntries = i ;

	i = 1 ;
	for( int n = 0 ; n < 5 ; n++ ) {
		for( int j = 0 ; j  < numCPEntries ; j++ ) {
			if( lst[j].level == n ) {
				CP[ i ] = lst[j] ;
				lst[j].index = (short)i++ ;
			}
		}
	}

	//
	// Now build the constructor (aka <init> method),
	// and the doActionMethod() which invokes the virtual
	// method that we're trying to reference...
	//
	MethodInfo initMethod = buildInitMethod();
	MethodInfo doActionMethod = buildDoActionMethod();

	//
	// build the sourcefile attr. Note: perhaps this can be a constant
	// check it out later...
	// 
	SourceFileAttr sourceFileAttr = new SourceFileAttr( SourceFile, javaFile ) ;	



	//-----------------------------------------------
	// Now write everything out to a DataOutputStream
	//-----------------------------------------------
		
	out.writeInt(JAVA_MAGIC );		// 0xCAFEBABE, ;-) !
	out.writeShort(JAVA_MINOR_VERSION);	// minor
	out.writeShort(JAVA_VERSION );	// major ver.

	out.writeShort( numCPEntries + 1);	// number of constant pool entries

	numCPEntries ++ ;
	for(  i = 1 ; i < numCPEntries ; i++ ) {
		CP[i].write( out ) ;		// write out all the constants
      }
	out.writeShort( ACC_PUBLIC ) ;	// class access flags
	out.writeShort( thisClass.index ) ;	// constant pool index to this class
	out.writeShort( superClass.index ) ;// ""

	out.writeShort( 0 ) ;			// number of interfaces
	out.writeShort( 0 ) ;			// number of fields in this class
	out.writeShort( 2 ) ;			// number of methods....
		
	initMethod.write( out ) ;		// write out init method
	doActionMethod.write( out ) ;		// write out the action method

	out.writeShort( 1 ) ;			// number of source file attribs.		
	sourceFileAttr.write( out ) ;		// finally, write out the src file attr
		
	
	byte classBytes[]  = f.toByteArray() ;	// extract the byte array

	//
	// Is class already in cache ?
	//
	Class aClass ;
	if( null != (aClass = (Class)cache.get(functorClassSignature)) ) {
		return aClass ;
	}

      if( !debug ) {
	//
	// Convert the byte array into a class
	//
	aClass = defineClass(classBytes, 0, classBytes.length) ;		

	//
	// Now resolve any references
	//
    	if( aClass != null )
             		resolveClass( aClass ) ;

	} else {
 
	   FileOutputStream f2 = new FileOutputStream("methodRefClasses/" +
                                                     functorClassName + ".class" ) ;
	   f2.write( classBytes ) ;
         aClass = Class.forName( functorClassSignature ) ;


	}
	//
	// Place it in the cache
	//
	if( aClass != null )
		cache.put( functorClassSignature , aClass ) ;

		
	return aClass ;					

    }

    //-------------------------------------------------
    // loadClass() - used internally in order to convert.
    // No need for loadClass, as wel already have it built,
    // and just need to define() and resolve it...
    // Note: due to undocumented features, this must return null...
    // Not for public consumption. Only valid when passed null . yuck
    //
    public synchronized Class loadClass( String name, boolean resolve )
						throws  ClassNotFoundException {

		echo( "VM attempting to load: " + name ) ;
		return null ;
    }



    //--------------------------------------------
    // flushCache() - Clears the cache of all Class
    // references...
    public synchronized void flushCache() {
	cache.clear() ;
    }


    //-------------------------------------------------
    // useCompiler() - Generates a class file via the javac
    // compiler. This produces a class file  exactly
    // the same as what's produced by the buildClassFile()
    // The intermediate java source is kept around
    // for illustrative purposes...
    // The source file and the resulting class file are placed in
    // cwd/methodRefClasses
    //
    Class useCompiler(	String packageString,
				String classString,
				String methodString,
				String functorClassSignature,
				String functorClassName,
				String paramClass,
				boolean overrideProtectStatus )
				throws Exception  {
				
		FileOutputStream f = null ;
		f = new FileOutputStream(	"methodRefClasses/" + 
							functorClassName +
							".java" ) ;
	
		BufferedOutputStream out = new BufferedOutputStream( f ) ;

		String s =
		"package methodRefClasses ;\r\n" +
		" \r\n" + "import " + 
		((packageString != null ) ?
			(packageString + ".*;\r\n") :
			(classString+ ";\r\n") )+
		"import sysExt.* ;\r\n" + 
		" \r\n" + 
		" \r\n" + 
		"public class " + functorClassName + " extends methodRef {\r\n" +

		"\tpublic " + functorClassName + "() { super() ; } ;\r\n" + 
		"\tpublic void doAction( Object _o_) {\r\n" + 
		"\t\t((" + classString + ")(O))." + methodString + "((" + 
											 paramClass +
											")_o_);\r\n" +
		"\t\treturn  ;\r\n" + 
		"\t}\r\n" +
		"}\r\n" ;

		byte list[] = new byte[ s.length() + 1 ] ;
		s.getBytes( 0, s.length() , list, 0 ) ; 
		try {
		out.write( list, 0, s.length() ) ;
		out.flush() ;
		} catch ( IOException e ) {
		System.out.println( "OOPS!! IOExceptiom 1!!!" ) ;
		}


		Process p = null ;
		p = Runtime.getRuntime().exec(
				"javac methodRefClasses\\" + functorClassName + ".java");


		p.waitFor();
		
		return Class.forName(functorClassSignature);

	}


    //--------------
    // buildInitMethod() - Builds the constructor for this class
    //
    MethodInfo	buildInitMethod() {
	//
	// Allocate and initialize linenumber attribute byte list..
	//
 	
	int initLineListLength = 6 ;
	byte initLineList[]  = new byte[ initLineListLength ] ;

	initLineList[0] = 0;	
	initLineList[1] = 1;	
	initLineList[2] = 0;	
	initLineList[3] = 0;	
	initLineList[4] = 0;	
	initLineList[5] = 8;

	//
	// Allocate and initialize byte codes for init method...
	//
      int initByteCodeLength = 5 ;
	byte initByteCodes[] = new byte [ initByteCodeLength ] ;

	initByteCodes[0] = (byte)opc_aload_0 ;
	initByteCodes[1] = (byte)opc_invokenonvirtual ;

	initByteCodes[2] = (byte)0 ;   
	initByteCodes[3] = (byte)initMethodConst.index ;
	initByteCodes[4] = (byte)opc_return ;

	// Now build the init method...
	//	
	return new MethodInfo(
			_init_,
			voidSig,
			new CodeAttr(
				Code,
				initByteCodeLength,
				initByteCodes, 
				new LineAttr(
					LineNumberTable,
					initLineListLength,
					initLineList
				),
				1, 1
			)
		) ;


    }


    //---------------------------------
    // buildDoActionMethod() - Builds the wrapper method which 
    // invokes the target method...
    //
    MethodInfo buildDoActionMethod() {

	//
	// Allocate and initialize linenumber attribute byte list..
	//
    	int doActionLineListLength = 10 ;
      byte doActionLineList[]  = new byte[ doActionLineListLength ] ;

	doActionLineList[0] = 0;	
	doActionLineList[1] = 2;	
	doActionLineList[2] = 0;	
	doActionLineList[3] = 0;	
	doActionLineList[4] = 0;	
	doActionLineList[5] = 10;	
	doActionLineList[6] = 0;	
	doActionLineList[7] = 11;	
	doActionLineList[8] = 0;
	doActionLineList[9] = 11;
	
	//
	// Allocate and initialize byte codes for doAction method...
	//
    	int	doActionByteCodeLength =  paramClassIsObject ? 12 : 15 ;
	byte doActionByteCodes[] = new byte [ doActionByteCodeLength ] ;
	int i = 0 ;

	//
	// Get the target object...
      //
	doActionByteCodes [i++] = (byte)opc_aload_0 ;
	doActionByteCodes [i++] = (byte)opc_getfield ;
	doActionByteCodes [i++] = (byte)0 ;
	doActionByteCodes [i++] = (byte)targetObj.index;

	//
	// ...and check that it can be cast to the target class
      //
	doActionByteCodes [i++] = (byte)opc_checkcast ;
	doActionByteCodes [i++] = (byte)0 ;
	doActionByteCodes [i++] = (byte)targetClass.index;
	
	//
	// Now get the parameter to be passed to
	// the target's action method..
	//	
	doActionByteCodes [i++] = (byte)opc_aload_1;

	//
	// If the parameter's class is not Object
	// then check that it's of the appropriate type
	//
	if( !paramClassIsObject ) {
      	doActionByteCodes [i++] = (byte)opc_checkcast ;
      	doActionByteCodes [i++] = (byte)0 ;
      	doActionByteCodes [i++] = (byte)paramClass.index;
	
		doActionLineList[7] = 14;// modify debugging line num info..	
			
	}


	//
	// invoke the target object's method...
	//
	doActionByteCodes [i++] = (byte)opc_invokevirtual ;
	doActionByteCodes [i++] = (byte)0 ;
	doActionByteCodes [i++]= (byte)actionMethodConst.index ;

	doActionByteCodes [i++]= (byte)opc_return ;

	//
	// now build the doAction method()...
	//
	return new MethodInfo(
			doAction,
			doActionSig,
				new CodeAttr(
					Code,
					doActionByteCodeLength,
					doActionByteCodes,
					new LineAttr(
						LineNumberTable,
						doActionLineListLength,
						doActionLineList
					),
					2, 2
				)
		);
	
    }


    void echo( String s ) { 
     // System.out.println( s ) ;
    } ;

} 
