package sysExt ;
/*
 * @(#)ConstAttr.java	1.0 96/01/05  Sal Cataudella
 *
 * Copyright (c) 1995 Sal Cataudella. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please contact author for other arrangements.
 *
 *
 */
import java.io.IOException;
import java.io.DataOutputStream;


//---------------------------------------------------------
// All classes in here represent constant pool entries....
//



//-------------------------------------------------------------
// ConstAttrInfo - Abstract class for all the Constant pool entries
//-------------------------------------------------------------
abstract class ConstAttrInfo implements RuntimeConstants {

	short index ;	// index in the Constant Pool table
	short level ;	// order in constant set 
	//---------------------------------
	// Writes itself out to whatever stream it is given
	// Must be implemented by every subclass...
	//
	abstract void write( DataOutputStream out ) throws IOException ;

} ;


//-------------------------------------------------------------
// ConstStringInfo - Represents constant string values 
//-------------------------------------------------------------
class ConstStringInfo extends ConstAttrInfo {
	byte bytes[] = null ;
	short length ;

	ConstStringInfo( String str ) {
		level = 4 ;
		//-----------------------------------------
		// Convert here to optimize for the statics
		//
		length = (short)str.length() ;
                bytes = new byte[ length ] ;
                str.getBytes( 0, length , bytes, 0 ) ;
	}

	
	void write( DataOutputStream out ) throws IOException {
               out.writeByte(CONSTANT_UTF8) ;
		out.writeShort( length ) ;
		out.write( bytes, 0, length ) ;
	}
	
} ;

//-------------------------------------------------------------
// ConstNameAndTypeInfo - Represents a variable or a method
// independant of any class
//-------------------------------------------------------------
class	ConstNameAndTypeInfo extends ConstAttrInfo {
	ConstStringInfo name ;
       ConstStringInfo signature ;

	ConstNameAndTypeInfo(	ConstStringInfo nameString,
                                   ConstStringInfo signatureString ) {
		level = 3 ;
            name = nameString;
            signature = signatureString;
	}

	void write( DataOutputStream out ) throws IOException {
		out.writeByte(CONSTANT_NAMEANDTYPE ) ;
              out.writeShort( name.index ) ;
		out.writeShort(signature.index ) ;
	}
}

//-------------------------------------------------------------
// ConstFieldInfo - Represents all facets of a variable 
// including which class.
//-------------------------------------------------------------
class ConstFieldInfo extends ConstAttrInfo {
	ConstClassInfo clazz ;
	ConstNameAndTypeInfo nameAndType ;

        ConstFieldInfo( ConstClassInfo classString,
				ConstNameAndTypeInfo nameAndTypeParm) {
		level = 2 ;
            clazz = classString;
		this.nameAndType = nameAndTypeParm;
	}

	void write( DataOutputStream out ) throws IOException {
		out.writeByte( CONSTANT_FIELD ) ;
		out.writeShort( clazz.index ) ;
		out.writeShort(nameAndType.index ) ;
	}
}
//-------------------------------------------------------------
// ConstMethodInfo - Same as ConstFieldInfo 
//-------------------------------------------------------------
class ConstMethodInfo extends ConstFieldInfo {
	ConstMethodInfo(ConstClassInfo classString,
			   ConstNameAndTypeInfo nameAndType){
		super(classString, nameAndType) ;
		level = 1 ;
	}

	void write( DataOutputStream out ) throws IOException {
		out.writeByte( CONSTANT_METHOD ) ;
		out.writeShort( clazz.index ) ;
		out.writeShort(nameAndType.index ) ;
	}


 }




//-------------------------------------------------------------
// ConstClassInfo - Represents constant classvalues 
//-------------------------------------------------------------
class ConstClassInfo extends ConstAttrInfo {
	ConstStringInfo name ;

        ConstClassInfo ( ConstStringInfo className) {
		level = 0 ;
		//-----------------------------------------
		// Convert here to optimize for the statics
		//
		name = className ;
	}

	
	void write( DataOutputStream out ) throws IOException {
		out.writeByte(CONSTANT_CLASS) ;
		out.writeShort( name.index ) ;
	}
	
} ;
